@@ -25,8 +25,9 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: yasm.c 1164 2004-10-14 05:55:21Z peter $");
+/*@unused@*/ RCSID("$Id: yasm.c 1591 2006-07-13 03:25:38Z peter $");
 
+#include <ctype.h>
 #include <libyasm/compat-queue.h>
 #include <libyasm/bitvect.h>
 #include <libyasm.h>
@@ -35,37 +36,15 @@
 #include <libgen.h>
 #endif
 
-#ifndef WIN32
-#include "ltdl.h"
-#endif
-#include "yasm-module.h"
 #include "yasm-options.h"
 
+#include "license.c"
 
-/* Extra path to search for our modules. */
-#ifndef YASM_MODULE_PATH_ENV
-# define YASM_MODULE_PATH_ENV	"YASM_MODULE_PATH"
-#endif
-
-#ifndef WIN32
-extern const lt_dlsymlist lt_preloaded_symbols[];
-#endif
+#define DEFAULT_OBJFMT_MODULE	"bin"
 
 /* Preprocess-only buffer size */
 #define PREPROC_BUF_SIZE    16384
 
-/* Check the module version */
-#define check_module_version(d, TYPE, type)	\
-do { \
-    if (d && d->version != YASM_##TYPE##_VERSION) { \
-	print_error( \
-	    _("%s: module version mismatch: %s `%s' (need %d, module %d)"), \
-	    _("FATAL"), _(#type), d->keyword, YASM_##TYPE##_VERSION, \
-	    d->version); \
-	exit(EXIT_FAILURE); \
-    } \
-} while (0)
-
 /*@null@*/ /*@only@*/ static char *obj_filename = NULL, *in_filename = NULL;
 /*@null@*/ /*@only@*/ static char *list_filename = NULL;
 /*@null@*/ /*@only@*/ static char *machine_name = NULL;
@@ -78,6 +57,7 @@ static int special_options = 0;
 /*@null@*/ /*@dependent@*/ static yasm_preproc *cur_preproc = NULL;
 /*@null@*/ /*@dependent@*/ static const yasm_preproc_module *
     cur_preproc_module = NULL;
+/*@null@*/ static char *objfmt_keyword = NULL;
 /*@null@*/ /*@dependent@*/ static yasm_objfmt *cur_objfmt = NULL;
 /*@null@*/ /*@dependent@*/ static const yasm_objfmt_module *
     cur_objfmt_module = NULL;
@@ -90,6 +70,7 @@ static int special_options = 0;
 /*@null@*/ /*@dependent@*/ static const yasm_listfmt_module *
     cur_listfmt_module = NULL;
 static int preproc_only = 0;
+static int generate_make_dependencies = 0;
 static int warning_error = 0;	/* warnings being treated as errors */
 static enum {
     EWSTYLE_GNU = 0,
@@ -114,7 +95,8 @@ static int opt_machine_handler(char *cmd, /*@null@*/ char *param, int extra);
 static int opt_warning_handler(char *cmd, /*@null@*/ char *param, int extra);
 static int preproc_only_handler(char *cmd, /*@null@*/ char *param, int extra);
 static int opt_preproc_option(char *cmd, /*@null@*/ char *param, int extra);
-static int opt_ewmsg_handler(char *cmd, /*@null@*/ char *param, int extra); 
+static int opt_ewmsg_handler(char *cmd, /*@null@*/ char *param, int extra);
+static int opt_makedep_handler(char *cmd, /*@null@*/ char *param, int extra);
 
 static /*@only@*/ char *replace_extension(const char *orig, /*@null@*/
 					  const char *ext, const char *def);
@@ -130,19 +112,23 @@ static void print_yasm_error(const char *filename, unsigned long line,
 static void print_yasm_warning(const char *filename, unsigned long line,
 			       const char *msg);
 
+static void apply_preproc_builtins(void);
 static void apply_preproc_saved_options(void);
 static void print_list_keyword_desc(const char *name, const char *keyword);
 
 /* values for special_options */
 #define SPECIAL_SHOW_HELP 0x01
 #define SPECIAL_SHOW_VERSION 0x02
-#define SPECIAL_LISTED 0x04
+#define SPECIAL_SHOW_LICENSE 0x04
+#define SPECIAL_LISTED 0x08
 
 /* command line options */
 static opt_option options[] =
 {
     { 0, "version", 0, opt_special_handler, SPECIAL_SHOW_VERSION,
       N_("show version text"), NULL },
+    { 0, "license", 0, opt_special_handler, SPECIAL_SHOW_LICENSE,
+      N_("show license text"), NULL },
     { 'h', "help", 0, opt_special_handler, SPECIAL_SHOW_HELP,
       N_("show help text"), NULL },
     { 'a', "arch", 1, opt_arch_handler, 0,
@@ -167,8 +153,12 @@ static opt_option options[] =
       N_("inhibits warning messages"), NULL },
     { 'W', NULL, 0, opt_warning_handler, 0,
       N_("enables/disables warning"), NULL },
+    { 'M', NULL, 0, opt_makedep_handler, 0,
+      N_("generate Makefile dependencies on stdout"), NULL },
     { 'e', "preproc-only", 0, preproc_only_handler, 0,
       N_("preprocess only (writes output to stdout by default)"), NULL },
+    { 'i', NULL, 1, opt_preproc_option, 0,
+      N_("add include path"), N_("path") },
     { 'I', NULL, 1, opt_preproc_option, 0,
       N_("add include path"), N_("path") },
     { 'P', NULL, 1, opt_preproc_option, 1,
@@ -183,35 +173,10 @@ static opt_option options[] =
 
 /* version message */
 /*@observer@*/ static const char *version_msg[] = {
-    PACKAGE " " VERSION "\n",
-    N_("Copyright (c) 2001-2003 Peter Johnson and other"), " " PACKAGE " ",
-    N_("developers.\n"),
-    N_("**Licensing summary**\n"),
-    N_("Note: This summary does not provide legal advice nor is it the\n"),
-    N_(" actual license.  See the individual licenses for complete\n"),
-    N_(" details.  Consult a laywer for legal advice.\n"),
-    N_("The primary license is the 2-clause BSD license.  Please use this\n"),
-    N_(" license if you plan on submitting code to the project.\n"),
-    N_("Libyasm:\n"),
-    N_(" Libyasm is 2-clause or 3-clause BSD licensed, with the exception\n"),
-    N_(" of bitvect, which is triple-licensed under the Artistic license,\n"),
-    N_(" GPL, and LGPL.  Libyasm is thus GPL and LGPL compatible.  In\n"),
-    N_(" addition, this also means that libyasm is free for binary-only\n"),
-    N_(" distribution as long as the terms of the 3-clause BSD license and\n"),
-    N_(" Artistic license (as it applies to bitvect) are fulfilled.\n"),
-    N_("Modules:\n"),
-    N_(" Most of the modules are 2-clause BSD licensed, except:\n"),
-    N_("  preprocs/nasm - LGPL licensed\n"),
-    N_("Frontends:\n"),
-    N_(" The frontends are 2-clause BSD licensed.\n"),
-    N_("License Texts:\n"),
-    N_(" The full text of all licenses are provided in separate files in\n"),
-    N_(" this program's source distribution.  Each file may include the\n"),
-    N_(" entire license (in the case of the BSD and Artistic licenses), or\n"),
-    N_(" may reference the GPL or LGPL license file.\n"),
-    N_("This program has absolutely no warranty; not even for\n"),
-    N_("merchantibility or fitness for a particular purpose.\n"),
-    N_("Compiled on"), " " __DATE__ ".\n",
+    PACKAGE_NAME " " PACKAGE_INTVER "." PACKAGE_BUILD,
+    "Compiled on " __DATE__ ".",
+    "Copyright (c) 2001-2006 Peter Johnson and other Yasm developers.",
+    "Run yasm --license for licensing overview and summary."
 };
 
 /* help messages */
@@ -247,9 +212,8 @@ main(int argc, char *argv[])
     yasm_object *object = NULL;
     yasm_section *def_sect;
     size_t i;
-#ifndef WIN32
-    int errors;
-#endif
+    yasm_arch_create_error arch_error;
+    const char *base_filename;
 
 #if defined(HAVE_SETLOCALE) && defined(HAVE_LC_MESSAGES)
     setlocale(LC_MESSAGES, "");
@@ -265,42 +229,6 @@ main(int argc, char *argv[])
     yasm_gettext_hook = handle_yasm_gettext;
     yasm_errwarn_initialize();
 
-#ifndef WIN32
-    /* Set libltdl malloc/free functions. */
-#ifdef WITH_DMALLOC
-    lt_dlmalloc = malloc;
-    lt_dlfree = free;
-#else
-    lt_dlmalloc = yasm_xmalloc;
-    lt_dlfree = yasm_xfree;
-#endif
-
-    /* Initialize preloaded symbol lookup table. */
-    lt_dlpreload_default(lt_preloaded_symbols);
-
-    /* Initialize libltdl. */
-    errors = lt_dlinit();
-
-    /* Set up extra module search directories. */
-    if (errors == 0) {
-	const char *path = getenv(YASM_MODULE_PATH_ENV);
-	if (path)
-	    errors = lt_dladdsearchdir(path);
-    }
-#if defined(YASM_MODULEDIR)
-    if (errors == 0)
-	errors = lt_dladdsearchdir(YASM_MODULEDIR);
-#endif
-    if (errors == 0) {
-	/* Path where yasm executable is running from (argv[0]) */
-	errors = lt_dladdsearchdir(dirname(argv[0]));
-    }
-    if (errors != 0) {
-	print_error(_("%s: module loader initialization failed"), _("FATAL"));
-	return EXIT_FAILURE;
-    }
-#endif
-
     /* Initialize parameter storage */
     STAILQ_INIT(&preproc_options);
 
@@ -313,8 +241,12 @@ main(int argc, char *argv[])
 	    help_msg(help_head, help_tail, options, NELEMS(options));
 	    return EXIT_SUCCESS;
 	case SPECIAL_SHOW_VERSION:
-	    for (i=0; i<sizeof(version_msg)/sizeof(char *); i++)
-		printf("%s", gettext(version_msg[i]));
+	    for (i=0; i<NELEMS(version_msg); i++)
+		printf("%s\n", version_msg[i]);
+	    return EXIT_SUCCESS;
+	case SPECIAL_SHOW_LICENSE:
+	    for (i=0; i<NELEMS(license_msg); i++)
+		printf("%s\n", license_msg[i]);
 	    return EXIT_SUCCESS;
 	case SPECIAL_LISTED:
 	    /* Printed out earlier */
@@ -348,6 +280,16 @@ main(int argc, char *argv[])
     /* Initialize intnum and floatnum */
     yasm_intnum_initialize();
     yasm_floatnum_initialize();
+    
+    /* If not already specified, default to bin as the object format. */
+    if (!cur_objfmt_module)
+	cur_objfmt_module = yasm_load_objfmt(DEFAULT_OBJFMT_MODULE);
+
+    if (!cur_objfmt_module) {
+	print_error(_("%s: could not load default %s"), _("FATAL"),
+		    _("object format"));
+	return EXIT_FAILURE;
+    }
 
     /* handle preproc-only case here */
     if (preproc_only) {
@@ -359,10 +301,28 @@ main(int argc, char *argv[])
 	linemap = yasm_linemap_create();
 	yasm_linemap_set(linemap, in_filename, 1, 1);
 
-	/* Default output to stdout if not specified */
-	if (!obj_filename)
+	/* Default output to stdout if not specified or generating dependency
+           makefiles */
+	if (!obj_filename || generate_make_dependencies) {
 	    obj = stdout;
-	else {
+
+            /* determine the object filename if not specified, but we need a
+               file name for the makefile rule */
+            if (generate_make_dependencies && !obj_filename) {
+                if (in == stdin)
+                    /* Default to yasm.out if no obj filename specified */
+                    obj_filename = yasm__xstrdup("yasm.out");
+                else {
+                    /* replace (or add) extension to base filename */
+		    yasm__splitpath(in_filename, &base_filename);
+		    if (base_filename[0] == '\0')
+			obj_filename = yasm__xstrdup("yasm.out");
+		    else
+			obj_filename = replace_extension(base_filename,
+                            cur_objfmt_module->extension, "yasm.out");
+		}
+            }
+	} else {
 	    /* Open output (object) file */
 	    obj = open_file(obj_filename, "wt");
 	    if (!obj) {
@@ -373,7 +333,7 @@ main(int argc, char *argv[])
 
 	/* If not already specified, default to nasm preproc. */
 	if (!cur_preproc_module)
-	    cur_preproc_module = load_preproc_module("nasm");
+	    cur_preproc_module = yasm_load_preproc("nasm");
 
 	if (!cur_preproc_module) {
 	    print_error(_("%s: could not load default %s"), _("FATAL"),
@@ -381,16 +341,36 @@ main(int argc, char *argv[])
 	    cleanup(NULL);
 	    return EXIT_FAILURE;
 	}
-	check_module_version(cur_preproc_module, PREPROC, preproc);
-
-	apply_preproc_saved_options();
-
+	
 	/* Pre-process until done */
 	cur_preproc = yasm_preproc_create(cur_preproc_module, in, in_filename,
 					  linemap);
-	while ((got = yasm_preproc_input(cur_preproc, preproc_buf,
-					 PREPROC_BUF_SIZE)) != 0)
-	    fwrite(preproc_buf, got, 1, obj);
+
+	apply_preproc_builtins();
+	apply_preproc_saved_options();
+
+        if (generate_make_dependencies) {
+	    size_t totlen;
+
+            fprintf(stdout, "%s: %s", obj_filename, in_filename);
+	    totlen = strlen(obj_filename)+2+strlen(in_filename);
+
+            while ((got = yasm_preproc_get_included_file(cur_preproc,
+				preproc_buf, PREPROC_BUF_SIZE)) != 0) {
+		totlen += got;
+		if (totlen > 72) {
+		    fputs(" \\\n  ", stdout);
+		    totlen = 2;
+		}
+		fputc(' ', stdout);
+                fwrite(preproc_buf, got, 1, stdout);
+	    }
+	    fputc('\n', stdout);
+        } else {
+            while ((got = yasm_preproc_input(cur_preproc, preproc_buf,
+					     PREPROC_BUF_SIZE)) != 0)
+	        fwrite(preproc_buf, got, 1, obj);
+        }
 
 	if (in != stdin)
 	    fclose(in);
@@ -414,26 +394,67 @@ main(int argc, char *argv[])
 	return EXIT_SUCCESS;
     }
 
+    /* determine the object filename if not specified */
+    if (!obj_filename) {
+	if (in == stdin)
+	    /* Default to yasm.out if no obj filename specified */
+	    obj_filename = yasm__xstrdup("yasm.out");
+	else {
+	    /* replace (or add) extension to base filename */
+	    yasm__splitpath(in_filename, &base_filename);
+	    if (base_filename[0] == '\0')
+		obj_filename = yasm__xstrdup("yasm.out");
+	    else
+		obj_filename = replace_extension(base_filename,
+						 cur_objfmt_module->extension,
+						 "yasm.out");
+	}
+    }
+
     /* Create object */
-    object = yasm_object_create();
+    object = yasm_object_create(in_filename, obj_filename);
     yasm_linemap_set(yasm_object_get_linemap(object), in_filename, 1, 1);
 
     /* Default to x86 as the architecture */
     if (!cur_arch_module) {
-	cur_arch_module = load_arch_module("x86");
+	cur_arch_module = yasm_load_arch("x86");
 	if (!cur_arch_module) {
 	    print_error(_("%s: could not load default %s"), _("FATAL"),
 			_("architecture"));
 	    return EXIT_FAILURE;
 	}
     }
-    check_module_version(cur_arch_module, ARCH, arch);
 
-    /* Set up architecture using the selected (or default) machine */
-    if (!machine_name)
-	machine_name = yasm__xstrdup(cur_arch_module->default_machine_keyword);
+    /* Default to NASM as the parser */
+    if (!cur_parser_module) {
+	cur_parser_module = yasm_load_parser("nasm");
+	if (!cur_parser_module) {
+	    print_error(_("%s: could not load default %s"), _("FATAL"),
+			_("parser"));
+	    cleanup(NULL);
+	    return EXIT_FAILURE;
+	}
+    }
 
-    cur_arch = cur_arch_module->create(machine_name);
+    /* Set up architecture using the selected (or default) machine and
+     * selected (or default nasm) parser.
+     */
+    if (!machine_name) {
+	/* If we're using x86 and the default objfmt bits is 64, default the
+	 * machine to amd64.  When we get more arches with multiple machines,
+	 * we should do this in a more modular fashion.
+	 */
+	if (strcmp(cur_arch_module->keyword, "x86") == 0 &&
+	    cur_objfmt_module->default_x86_mode_bits == 64)
+	    machine_name = yasm__xstrdup("amd64");
+	else
+	    machine_name =
+		yasm__xstrdup(cur_arch_module->default_machine_keyword);
+    }
+
+    cur_arch = cur_arch_module->create(machine_name,
+				       cur_parser_module->keyword,
+				       &arch_error);
     if (!cur_arch) {
 	if (strcmp(machine_name, "help") == 0) {
 	    yasm_arch_machine *m = cur_arch_module->machines;
@@ -446,43 +467,43 @@ main(int argc, char *argv[])
 	    return EXIT_SUCCESS;
 	}
 
-	print_error(_("%s: `%s' is not a valid %s for %s `%s'"),
-		    _("FATAL"), machine_name, _("machine"),
-		    _("architecture"), cur_arch_module->keyword);
+	switch (arch_error) {
+	    case YASM_ARCH_CREATE_BAD_MACHINE:
+		print_error(_("%s: `%s' is not a valid %s for %s `%s'"),
+			    _("FATAL"), machine_name, _("machine"),
+			    _("architecture"), cur_arch_module->keyword);
+		break;
+	    case YASM_ARCH_CREATE_BAD_PARSER:
+		print_error(_("%s: `%s' is not a valid %s for %s `%s'"),
+			    _("FATAL"), machine_name, _("machine"),
+			    _("architecture"), cur_arch_module->keyword);
+		break;
+	    default:
+		print_error(_("%s: unknown architecture error"), _("FATAL"));
+	}
+
 	return EXIT_FAILURE;
     }
 
     /* Set basic as the optimizer (TODO: user choice) */
-    cur_optimizer_module = load_optimizer_module("basic");
+    cur_optimizer_module = yasm_load_optimizer("basic");
 
     if (!cur_optimizer_module) {
 	print_error(_("%s: could not load default %s"), _("FATAL"),
 		    _("optimizer"));
 	return EXIT_FAILURE;
     }
-    check_module_version(cur_optimizer_module, OPTIMIZER, optimizer);
 
     /* If list file enabled, make sure we have a list format loaded. */
     if (list_filename) {
 	/* If not already specified, default to nasm as the list format. */
 	if (!cur_listfmt_module)
-	    cur_listfmt_module = load_listfmt_module("nasm");
+	    cur_listfmt_module = yasm_load_listfmt("nasm");
     }
 
-    /* If not already specified, default to bin as the object format. */
-    if (!cur_objfmt_module)
-	cur_objfmt_module = load_objfmt_module("bin");
-
-    if (!cur_objfmt_module) {
-	print_error(_("%s: could not load default %s"), _("FATAL"),
-		    _("object format"));
-	return EXIT_FAILURE;
-    }
-    check_module_version(cur_objfmt_module, OBJFMT, objfmt);
-
     /* If not already specified, default to null as the debug format. */
     if (!cur_dbgfmt_module)
-	cur_dbgfmt_module = load_dbgfmt_module("null");
+	cur_dbgfmt_module = yasm_load_dbgfmt("null");
     else {
 	int matched_dbgfmt = 0;
 	/* Check to see if the requested debug format is in the allowed list
@@ -508,23 +529,9 @@ main(int argc, char *argv[])
 		    _("debug format"));
 	return EXIT_FAILURE;
     }
-    check_module_version(cur_dbgfmt_module, DBGFMT, dbgfmt);
-
-    /* determine the object filename if not specified */
-    if (!obj_filename) {
-	if (in == stdin)
-	    /* Default to yasm.out if no obj filename specified */
-	    obj_filename = yasm__xstrdup("yasm.out");
-	else
-	    /* replace (or add) extension */
-	    obj_filename = replace_extension(in_filename,
-					     cur_objfmt_module->extension,
-					     "yasm.out");
-    }
 
     /* Initialize the object format */
-    cur_objfmt = yasm_objfmt_create(cur_objfmt_module, in_filename, object,
-				    cur_arch);
+    cur_objfmt = yasm_objfmt_create(cur_objfmt_module, object, cur_arch);
     if (!cur_objfmt) {
 	print_error(
 	    _("%s: object format `%s' does not support architecture `%s' machine `%s'"),
@@ -535,12 +542,14 @@ main(int argc, char *argv[])
 	return EXIT_FAILURE;
     }
 
+    /* Get a fresh copy of objfmt_module as it may have changed. */
+    cur_objfmt_module = ((yasm_objfmt_base *)cur_objfmt)->module;
+
     /* Add an initial "default" section to object */
-    def_sect = yasm_objfmt_add_default_section(cur_objfmt, object);
+    def_sect = yasm_objfmt_add_default_section(cur_objfmt);
 
     /* Initialize the debug format */
-    cur_dbgfmt = yasm_dbgfmt_create(cur_dbgfmt_module, in_filename,
-				    obj_filename, object, cur_objfmt,
+    cur_dbgfmt = yasm_dbgfmt_create(cur_dbgfmt_module, object, cur_objfmt,
 				    cur_arch);
     if (!cur_dbgfmt) {
 	print_error(
@@ -552,22 +561,10 @@ main(int argc, char *argv[])
 	return EXIT_FAILURE;
     }
 
-    /* Default to NASM as the parser */
-    if (!cur_parser_module) {
-	cur_parser_module = load_parser_module("nasm");
-	if (!cur_parser_module) {
-	    print_error(_("%s: could not load default %s"), _("FATAL"),
-			_("parser"));
-	    cleanup(NULL);
-	    return EXIT_FAILURE;
-	}
-    }
-    check_module_version(cur_parser_module, PARSER, parser);
-
     /* If not already specified, default to the parser's default preproc. */
     if (!cur_preproc_module)
 	cur_preproc_module =
-	    load_preproc_module(cur_parser_module->default_preproc_keyword);
+	    yasm_load_preproc(cur_parser_module->default_preproc_keyword);
     else {
 	int matched_preproc = 0;
 	/* Check to see if the requested preprocessor is in the allowed list
@@ -595,11 +592,11 @@ main(int argc, char *argv[])
 	cleanup(NULL);
 	return EXIT_FAILURE;
     }
-    check_module_version(cur_preproc_module, PREPROC, preproc);
 
     cur_preproc = cur_preproc_module->create(in, in_filename,
 					     yasm_object_get_linemap(object));
 
+    apply_preproc_builtins();
     apply_preproc_saved_options();
 
     /* Get initial x86 BITS setting from object format */
@@ -609,13 +606,19 @@ main(int argc, char *argv[])
     }
 
     /* Parse! */
-    cur_parser_module->do_parse(object, cur_preproc, cur_arch, cur_objfmt, in,
-				in_filename, list_filename != NULL, def_sect);
+    cur_parser_module->do_parse(object, cur_preproc, cur_arch, cur_objfmt,
+				cur_dbgfmt, in, in_filename,
+				list_filename != NULL, def_sect);
 
     /* Close input file */
     if (in != stdin)
 	fclose(in);
 
+    /* Check for undefined symbols */
+    yasm_symtab_parser_finalize(yasm_object_get_symtab(object),
+				strcmp(cur_parser_module->keyword, "gas")==0,
+				cur_objfmt);
+
     if (yasm_get_num_errors(warning_error) > 0) {
 	yasm_errwarn_output_all(yasm_object_get_linemap(object), warning_error,
 				print_yasm_error, print_yasm_warning);
@@ -623,7 +626,17 @@ main(int argc, char *argv[])
 	return EXIT_FAILURE;
     }
 
-    yasm_symtab_parser_finalize(yasm_object_get_symtab(object));
+    /* Finalize parse */
+    yasm_object_finalize(object);
+
+    if (yasm_get_num_errors(warning_error) > 0) {
+	yasm_errwarn_output_all(yasm_object_get_linemap(object), warning_error,
+				print_yasm_error, print_yasm_warning);
+	cleanup(object);
+	return EXIT_FAILURE;
+    }
+
+    /* Optimize */
     cur_optimizer_module->optimize(object);
 
     if (yasm_get_num_errors(warning_error) > 0) {
@@ -646,7 +659,7 @@ main(int argc, char *argv[])
     }
 
     /* Write the object file */
-    yasm_objfmt_output(cur_objfmt, obj?obj:stderr, obj_filename,
+    yasm_objfmt_output(cur_objfmt, obj?obj:stderr,
 		       strcmp(cur_dbgfmt_module->keyword, "null"), cur_dbgfmt);
 
     /* Close object file */
@@ -731,13 +744,6 @@ cleanup(yasm_object *object)
 	BitVector_Shutdown();
     }
 
-    unload_modules();
-
-#ifndef WIN32
-    /* Finish with libltdl. */
-    lt_dlexit();
-#endif
-
     if (DO_FREE) {
 	if (in_filename)
 	    yasm_xfree(in_filename);
@@ -747,6 +753,8 @@ cleanup(yasm_object *object)
 	    yasm_xfree(list_filename);
 	if (machine_name)
 	    yasm_xfree(machine_name);
+	if (objfmt_keyword)
+	    yasm_xfree(objfmt_keyword);
     }
 }
 
@@ -779,11 +787,11 @@ static int
 opt_arch_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 {
     assert(param != NULL);
-    cur_arch_module = load_arch_module(param);
+    cur_arch_module = yasm_load_arch(param);
     if (!cur_arch_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("architectures"));
-	    list_archs(print_list_keyword_desc);
+	    yasm_list_arch(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -798,11 +806,11 @@ static int
 opt_parser_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 {
     assert(param != NULL);
-    cur_parser_module = load_parser_module(param);
+    cur_parser_module = yasm_load_parser(param);
     if (!cur_parser_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("parsers"));
-	    list_parsers(print_list_keyword_desc);
+	    yasm_list_parser(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -817,11 +825,11 @@ static int
 opt_preproc_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 {
     assert(param != NULL);
-    cur_preproc_module = load_preproc_module(param);
+    cur_preproc_module = yasm_load_preproc(param);
     if (!cur_preproc_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("preprocessors"));
-	    list_preprocs(print_list_keyword_desc);
+	    yasm_list_preproc(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -835,12 +843,13 @@ opt_preproc_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 static int
 opt_objfmt_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 {
+    size_t i;
     assert(param != NULL);
-    cur_objfmt_module = load_objfmt_module(param);
+    cur_objfmt_module = yasm_load_objfmt(param);
     if (!cur_objfmt_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("object formats"));
-	    list_objfmts(print_list_keyword_desc);
+	    yasm_list_objfmt(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -848,6 +857,11 @@ opt_objfmt_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 		    _("object format"), param);
 	exit(EXIT_FAILURE);
     }
+    if (objfmt_keyword)
+	yasm_xfree(objfmt_keyword);
+    objfmt_keyword = yasm__xstrdup(param);
+    for (i=0; i<strlen(objfmt_keyword); i++)
+	objfmt_keyword[i] = tolower(objfmt_keyword[i]);
     return 0;
 }
 
@@ -855,11 +869,11 @@ static int
 opt_dbgfmt_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
 {
     assert(param != NULL);
-    cur_dbgfmt_module = load_dbgfmt_module(param);
+    cur_dbgfmt_module = yasm_load_dbgfmt(param);
     if (!cur_dbgfmt_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("debug formats"));
-	    list_dbgfmts(print_list_keyword_desc);
+	    yasm_list_dbgfmt(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -875,11 +889,11 @@ opt_listfmt_handler(/*@unused@*/ char *cmd, char *param,
 		    /*@unused@*/ int extra)
 {
     assert(param != NULL);
-    cur_listfmt_module = load_listfmt_module(param);
+    cur_listfmt_module = yasm_load_listfmt(param);
     if (!cur_listfmt_module) {
 	if (!strcmp("help", param)) {
 	    printf(_("Available yasm %s:\n"), _("list formats"));
-	    list_listfmts(print_list_keyword_desc);
+	    yasm_list_listfmt(print_list_keyword_desc);
 	    special_options = SPECIAL_LISTED;
 	    return 0;
 	}
@@ -938,7 +952,8 @@ opt_machine_handler(/*@unused@*/ char *cmd, char *param,
 static int
 opt_warning_handler(char *cmd, /*@unused@*/ char *param, int extra)
 {
-    int enable = 1;	/* is it disabling the warning instead of enabling? */
+    /* is it disabling the warning instead of enabling? */
+    void (*action)(yasm_warn_class wclass) = yasm_warn_enable;
 
     if (extra == 1) {
 	/* -w, disable warnings */
@@ -951,26 +966,22 @@ opt_warning_handler(char *cmd, /*@unused@*/ char *param, int extra)
 
     /* detect no- prefix to disable the warning */
     if (cmd[0] == 'n' && cmd[1] == 'o' && cmd[2] == '-') {
-	enable = 0;
+	action = yasm_warn_disable;
 	cmd += 3;   /* skip past it to get to the warning name */
     }
 
     if (cmd[0] == '\0')
 	/* just -W or -Wno-, so definitely not valid */
 	return 1;
-    else if (strcmp(cmd, "error") == 0) {
-	warning_error = enable;
-    } else if (strcmp(cmd, "unrecognized-char") == 0) {
-	if (enable)
-	    yasm_warn_enable(YASM_WARN_UNREC_CHAR);
-	else
-	    yasm_warn_disable(YASM_WARN_UNREC_CHAR);
-    } else if (strcmp(cmd, "orphan-labels") == 0) {
-	if (enable)
-	    yasm_warn_enable(YASM_WARN_ORPHAN_LABEL);
-	else
-	    yasm_warn_disable(YASM_WARN_ORPHAN_LABEL);
-    } else
+    else if (strcmp(cmd, "error") == 0)
+	warning_error = (action == yasm_warn_enable);
+    else if (strcmp(cmd, "unrecognized-char") == 0)
+	action(YASM_WARN_UNREC_CHAR);
+    else if (strcmp(cmd, "orphan-labels") == 0)
+	action(YASM_WARN_ORPHAN_LABEL);
+    else if (strcmp(cmd, "uninit-contents") == 0)
+	action(YASM_WARN_UNINIT_CONTENTS);
+    else
 	return 1;
 
     return 0;
@@ -1009,6 +1020,34 @@ opt_ewmsg_handler(/*@unused@*/ char *cmd, char *param, /*@unused@*/ int extra)
     return 0;
 }
 
+static int
+opt_makedep_handler(/*@unused@*/ char *cmd, /*@unused@*/ char *param,
+		    /*@unused@*/ int extra)
+{
+    /* Also set preproc_only to 1, we don't want to generate code */
+    preproc_only = 1;
+    generate_make_dependencies = 1;
+    
+    return 0;
+}
+
+static void
+apply_preproc_builtins()
+{
+    char *predef;
+
+    if (!objfmt_keyword)
+	objfmt_keyword = yasm__xstrdup(DEFAULT_OBJFMT_MODULE);
+
+    /* Define standard YASM assembly-time macro constants */
+    predef = yasm_xmalloc(strlen("__YASM_OBJFMT__=")
+			  + strlen(objfmt_keyword) + 1);
+    strcpy(predef, "__YASM_OBJFMT__=");
+    strcat(predef, objfmt_keyword);
+    yasm_preproc_define_builtin(cur_preproc, predef);
+    yasm_xfree(predef);
+}
+
 static void
 apply_preproc_saved_options()
 {
