@@ -3,7 +3,7 @@
  * \brief YASM architecture interface.
  *
  * \rcs
- * $Id: arch.h 1168 2004-10-31 01:07:52Z peter $
+ * $Id: arch.h 1591 2006-07-13 03:25:38Z peter $
  * \endrcs
  *
  * \license
@@ -34,22 +34,33 @@
 #ifndef YASM_ARCH_H
 #define YASM_ARCH_H
 
-/** Return value from yasm_arch_module::parse_check_id(). */
+/** Errors that may be returned by yasm_arch_module::create(). */
 typedef enum {
-    YASM_ARCH_CHECK_ID_NONE = 0,	/**< Just a normal identifier. */
-    YASM_ARCH_CHECK_ID_INSN,		/**< An instruction. */
-    YASM_ARCH_CHECK_ID_PREFIX,		/**< An instruction prefix. */
-    YASM_ARCH_CHECK_ID_REG,		/**< A register. */
-    YASM_ARCH_CHECK_ID_SEGREG,/**< a segment register (for memory overrides). */
-    YASM_ARCH_CHECK_ID_TARGETMOD	/**< A target modifier (for jumps) */
-} yasm_arch_check_id_retval;
+    YASM_ARCH_CREATE_OK = 0,		/**< No error. */
+    YASM_ARCH_CREATE_BAD_MACHINE,	/**< Unrecognized machine name. */
+    YASM_ARCH_CREATE_BAD_PARSER		/**< Unrecognized parser name. */
+} yasm_arch_create_error;
+
+/** Return values for yasm_arch_module::parse_check_insnprefix(). */
+typedef enum {
+    YASM_ARCH_NOTINSNPREFIX = 0,	/**< Unrecognized */
+    YASM_ARCH_INSN,			/**< An instruction */
+    YASM_ARCH_PREFIX			/**< An instruction prefix */
+} yasm_arch_insnprefix;
+
+/** Types of registers / target modifiers that may be returned by
+ * yasm_arch_module::parse_check_regtmod().
+ */
+typedef enum {
+    YASM_ARCH_NOTREGTMOD = 0,		/**< Unrecognized */
+    YASM_ARCH_REG,			/**< A "normal" register */
+    YASM_ARCH_REGGROUP,			/**< A group of indexable registers */
+    YASM_ARCH_SEGREG,			/**< A segment register */
+    YASM_ARCH_TARGETMOD			/**< A target modifier (for jumps) */
+} yasm_arch_regtmod;
 
 /** An instruction operand (opaque type). */
 typedef struct yasm_insn_operand yasm_insn_operand;
-/** A list of instruction operands (opaque type).
- * The list goes from left-to-right as parsed.
- */
-typedef struct yasm_insn_operands yasm_insn_operands;
 #ifdef YASM_LIB_INTERNAL
 /*@reldef@*/ STAILQ_HEAD(yasm_insn_operands, yasm_insn_operand);
 #endif
@@ -64,22 +75,6 @@ typedef struct yasm_arch_base {
 } yasm_arch_base;
 #endif
 
-/** "Flavor" of the parser.
- * Different assemblers order instruction operands differently.  Also, some
- * differ on how exactly various registers are specified.  There's no great
- * solution to this, as the parsers aren't supposed to have knowledge of the
- * architectural internals, and the architecture is supposed to be parser
- * independent.  To make things work, as a rather hackish solution, we give the
- * architecture a little knowledge about the general "flavor" of the parser,
- * and let the architecture decide what to do with it.  Most architectures will
- * probably not even use this, but it's required for some (x86 in particular)
- * for correct behavior on all parsers.
- */
-typedef enum {
-    YASM_ARCH_SYNTAX_FLAVOR_NASM = 1,	/**< Like NASM */
-    YASM_ARCH_SYNTAX_FLAVOR_GAS		/**< Like GAS */
-} yasm_arch_syntax_flavor;
-
 /** YASM machine subtype.  A number of different machine types may be
  * associated with a single architecture.  These may be specific CPU's, but
  * the ABI used to interface with the architecture should be the primary
@@ -94,18 +89,6 @@ typedef struct yasm_arch_machine {
     const char *keyword;
 } yasm_arch_machine;
 
-/** Version number of #yasm_arch_module interface.  Any functional change to
- * the #yasm_arch_module interface should simultaneously increment this number.
- * This version should be checked by #yasm_arch_module loaders (using
- * yasm_arch_module_get_version()) to verify that the expected version (the
- * version defined by its libyasm header files) matches the loaded module
- * version (the version defined by the module's libyasm header files).  Doing
- * this will ensure that the module version's function definitions match the
- * module loader's function definitions.  The version number must never be
- * decreased.
- */
-#define YASM_ARCH_VERSION	3
-
 /** YASM architecture module interface.
  * \note All "data" in parser-related functions (yasm_arch_parse_*) needs to
  *	 start the parse initialized to 0 to make it okay for a parser-related
@@ -113,12 +96,6 @@ typedef struct yasm_arch_machine {
  *	 called before on the same piece of data.
  */
 typedef struct yasm_arch_module {
-    /** Version (see #YASM_ARCH_VERSION).  Should always be set to
-     * #YASM_ARCH_VERSION by the module source and checked against
-     * #YASM_ARCH_VERSION by the module loader.
-     */
-    unsigned int version;
-
     /** One-line description of the architecture.
      * Call yasm_arch_name() to get the name of a particular #yasm_arch.
      */
@@ -132,11 +109,9 @@ typedef struct yasm_arch_module {
     /** Create architecture.
      * Module-level implementation of yasm_arch_create().
      * Call yasm_arch_create() instead of calling this function.
-     * \param machine	keyword of machine in use (must be one listed in
-     *			#machines)
-     * \return NULL if machine not recognized.
      */
-    /*@only@*/ yasm_arch * (*create) (const char *machine);
+    /*@only@*/ yasm_arch * (*create) (const char *machine, const char *parser,
+				      yasm_arch_create_error *error);
 
     /** Module-level implementation of yasm_arch_destroy().
      * Call yasm_arch_destroy() instead of calling this function.
@@ -148,6 +123,11 @@ typedef struct yasm_arch_module {
      */
     const char * (*get_machine) (const yasm_arch *arch);
 
+    /** Module-level implementation of yasm_arch_get_address_size().
+     * Call yasm_arch_get_address_size() instead of calling this function.
+     */
+    unsigned int (*get_address_size) (const yasm_arch *arch);
+
     /** Module-level implementation of yasm_arch_set_var().
      * Call yasm_arch_set_var() instead of calling this function.
      */
@@ -156,15 +136,22 @@ typedef struct yasm_arch_module {
     /** Module-level implementation of yasm_arch_parse_cpu().
      * Call yasm_arch_parse_cpu() instead of calling this function.
      */
-    void (*parse_cpu) (yasm_arch *arch, const char *cpuid,
+    void (*parse_cpu) (yasm_arch *arch, const char *cpuid, size_t cpuid_len,
 		       unsigned long line);
 
-    /** Module-level implementation of yasm_arch_parse_check_id().
-     * Call yasm_arch_parse_check_id() instead of calling this function.
+    /** Module-level implementation of yasm_arch_parse_check_insnprefix().
+     * Call yasm_arch_parse_check_insnprefix() instead of calling this function.
      */
-    yasm_arch_check_id_retval (*parse_check_id)
-	(yasm_arch *arch, unsigned long data[4], const char *id,
-	 unsigned long line);
+    yasm_arch_insnprefix (*parse_check_insnprefix)
+	(yasm_arch *arch, /*@out@*/ unsigned long data[4], const char *id,
+	 size_t id_len, unsigned long line);
+
+    /** Module-level implementation of yasm_arch_parse_check_regtmod().
+     * Call yasm_arch_parse_check_regtmod() instead of calling this function.
+     */
+    yasm_arch_regtmod (*parse_check_regtmod)
+	(yasm_arch *arch, /*@out@*/ unsigned long *data, const char *id,
+	 size_t id_len, unsigned long line);
 
     /** Module-level implementation of yasm_arch_parse_directive().
      * Call yasm_arch_parse_directive() instead of calling this function.
@@ -174,31 +161,20 @@ typedef struct yasm_arch_module {
 			    /*@null@*/ yasm_valparamhead *objext_valparams,
 			    yasm_object *object, unsigned long line);
 
-    /** Module-level implementation of yasm_arch_parse_insn().
-     * Call yasm_arch_parse_insn() instead of calling this function.
+    /** Module-level implementation of yasm_arch_get_fill().
+     * Call yasm_arch_get_fill() instead of calling this function.
      */
-    /*@null@*/ yasm_bytecode * (*parse_insn)
-	(yasm_arch *arch, const unsigned long data[4], int num_operands,
-	 /*@null@*/ yasm_insn_operands *operands, yasm_bytecode *prev_bc,
-	 unsigned long line);
+    const unsigned char ** (*get_fill) (const yasm_arch *arch);
 
-    /** Module-level implementation of yasm_arch_parse_prefix().
-     * Call yasm_arch_parse_prefix() instead of calling this function.
+    /** Module-level implementation of yasm_arch_finalize_insn().
+     * Call yasm_arch_finalize_insn() instead of calling this function.
      */
-    void (*parse_prefix) (yasm_arch *arch, yasm_bytecode *bc,
-			  const unsigned long data[4], unsigned long line);
-
-    /** Module-level implementation of yasm_arch_parse_seg_prefix().
-     * Call yasm_arch_parse_seg_prefix() instead of calling this function.
-     */
-    void (*parse_seg_prefix) (yasm_arch *arch, yasm_bytecode *bc,
-			      unsigned long segreg, unsigned long line);
-
-    /** Module-level implementation of yasm_arch_parse_seg_override().
-     * Call yasm_arch_parse_seg_override() instead of calling this function.
-     */
-    void (*parse_seg_override) (yasm_arch *arch, yasm_effaddr *ea,
-				unsigned long segreg, unsigned long line);
+    void (*finalize_insn)
+	(yasm_arch *arch, yasm_bytecode *bc, yasm_bytecode *prev_bc,
+	 const unsigned long data[4], int num_operands,
+	 /*@null@*/ yasm_insn_operands *operands, int num_prefixes,
+	 unsigned long **prefixes, int num_segregs,
+	 const unsigned long *segregs);
 
     /** Module-level implementation of yasm_arch_floatnum_tobytes().
      * Call yasm_arch_floatnum_tobytes() instead of calling this function.
@@ -208,13 +184,6 @@ typedef struct yasm_arch_module {
 			     size_t valsize, size_t shift, int warn,
 			     unsigned long line);
 
-    /** Module-level implementation of yasm_arch_intnum_fixup_rel().
-     * Call yasm_arch_intnum_fixup_rel() instead of calling this function.
-     */
-    int (*intnum_fixup_rel) (yasm_arch *arch, yasm_intnum *intn,
-			     size_t valsize, const yasm_bytecode *bc,
-			     unsigned long line);
-
     /** Module-level implementation of yasm_arch_intnum_tobytes().
      * Call yasm_arch_intnum_tobytes() instead of calling this function.
      */
@@ -228,6 +197,12 @@ typedef struct yasm_arch_module {
      */
     unsigned int (*get_reg_size) (yasm_arch *arch, unsigned long reg);
 
+    /** Module-level implementation of yasm_arch_reggroup_get_reg().
+     * Call yasm_arch_reggroup_get_reg() instead of calling this function.
+     */
+    unsigned long (*reggroup_get_reg) (yasm_arch *arch, unsigned long reggroup,
+				       unsigned long regindex);
+
     /** Module-level implementation of yasm_arch_reg_print().
      * Call yasm_arch_reg_print() instead of calling this function.
      */
@@ -260,6 +235,12 @@ typedef struct yasm_arch_module {
      * #yasm_arch.
      */
     unsigned int wordsize;
+
+    /** Worst case minimum instruction length in bytes.
+     * Call yasm_arch_min_insn_len() to get the minimum instruction length of
+     * a particular #yasm_arch.
+     */
+    unsigned int min_insn_len;
 } yasm_arch_module;
 
 #ifdef YASM_LIB_INTERNAL
@@ -288,6 +269,16 @@ struct yasm_insn_operand {
 
     /** Specified size of the operand, in bytes.  0 if not user-specified. */
     unsigned int size;
+
+    /** Nonzero if dereference.  Used for "*foo" in GAS.
+     * The reason for this is that by default in GAS, an unprefixed value
+     * is a memory address, except for jumps/calls, in which case it needs a
+     * "*" prefix to become a memory address (otherwise it's an immediate).
+     * This isn't knowable in the parser stage, so the parser sets this flag
+     * to indicate the "*" prefix has been used, and the arch needs to adjust
+     * the operand type appropriately depending on the instruction type.
+     */
+    int deref;
 };
 #endif
 
@@ -309,14 +300,24 @@ const char *yasm_arch_keyword(const yasm_arch *arch);
  */
 unsigned int yasm_arch_wordsize(const yasm_arch *arch);
 
+/** Get the minimum instruction length of an architecture.
+ * \param arch	    architecture
+ * \return Minimum instruction length (in bytes).
+ */
+unsigned int yasm_arch_min_insn_len(const yasm_arch *arch);
+
 /** Create architecture.
  * \param module	architecture module
  * \param machine	keyword of machine in use (must be one listed in
- *			#machines)
- * \return NULL if machine not recognized, otherwise new architecture.
+ *			#yasm_arch_module.machines)
+ * \param parser	keyword of parser in use
+ * \param error		error return value
+ * \return NULL on error (error returned in error parameter), otherwise new
+ *	   architecture.
  */
 /*@only@*/ yasm_arch *yasm_arch_create(const yasm_arch_module *module,
-				       const char *machine);
+				       const char *machine, const char *parser,
+				       /*@out@*/ yasm_arch_create_error *error);
 
 /** Clean up, free any architecture-allocated memory.
  * \param arch	architecture
@@ -329,6 +330,12 @@ void yasm_arch_destroy(/*@only@*/ yasm_arch *arch);
  */
 const char *yasm_arch_get_machine(const yasm_arch *arch);
 
+/** Get architecture's active address size, in bits.
+ * \param arch	architecture
+ * \return Active address size (in bits).
+ */
+unsigned int yasm_arch_get_address_size(const yasm_arch *arch);
+
 /** Set any arch-specific variables.  For example, "mode_bits" in x86.
  * \param arch	architecture
  * \param var	variable name
@@ -340,32 +347,47 @@ int yasm_arch_set_var(yasm_arch *arch, const char *var, unsigned long val);
 /** Switch available instructions/registers/etc based on a user-specified
  * CPU identifier.  Should modify behavior ONLY of parse_* functions!  The
  * bytecode and output functions should be able to handle any CPU.
- * \param arch	architecture
- * \param cpuid	cpu identifier as in the input file
- * \param line	virtual line (from yasm_linemap)
+ * \param arch		architecture
+ * \param cpuid		cpu identifier as in the input file
+ * \param cpuid_len	length of cpu identifier string
+ * \param line		virtual line (from yasm_linemap)
  */
-void yasm_arch_parse_cpu(yasm_arch *arch, const char *cpuid,
+void yasm_arch_parse_cpu(yasm_arch *arch, const char *cpuid, size_t cpuid_len,
 			 unsigned long line);
 
 /** Check an generic identifier to see if it matches architecture specific
- * names for instructions, registers, etc.  Unrecognized identifiers should
- * be returned as #YASM_ARCH_CHECK_ID_NONE so they can be treated as normal
- * symbols.  Any additional data beyond just the type (almost always
- * necessary) should be returned into the space provided by the data
- * parameter.
- * \note Even though this is passed a data[4], only data[0] should be used
- *	 for #YASM_ARCH_CHECK_ID_TARGETMOD, #YASM_ARCH_CHECK_ID_REG, and
- *	 #YASM_ARCH_CHECK_ID_SEGREG return values.
- * \param arch	architecture
- * \param data	extra identification information (yasm_arch-specific)
- *		[output]
- * \param id	identifier as in the input file
- * \param line	virtual line (from yasm_linemap)
- * \return General type of identifier (#yasm_arch_check_id_retval).
+ * names for instructions or instruction prefixes.  Unrecognized identifiers
+ * should return #YASM_ARCH_NOTINSNPREFIX so they can be treated as normal
+ * symbols.  Any additional data beyond just the type (almost always necessary)
+ * should be returned into the space provided by the data parameter.
+ * \param arch		architecture
+ * \param data		extra identification information (yasm_arch-specific)
+ *			[output]
+ * \param id		identifier as in the input file
+ * \param id_len	length of id string
+ * \param line		virtual line (from yasm_linemap)
+ * \return Identifier type (#YASM_ARCH_NOTINSNPREFIX if unrecognized)
  */
-yasm_arch_check_id_retval yasm_arch_parse_check_id
-    (yasm_arch *arch, unsigned long data[4], const char *id,
-     unsigned long line);
+yasm_arch_insnprefix yasm_arch_parse_check_insnprefix
+    (yasm_arch *arch, /*@out@*/ unsigned long data[4], const char *id,
+     size_t id_len, unsigned long line);
+
+/** Check an generic identifier to see if it matches architecture specific
+ * names for registers or target modifiers.  Unrecognized identifiers should
+ * return #YASM_ARCH_NOTREGTMOD.  Any additional data beyond just the type
+ * (almost always necessary) should be returned into the space provided by the
+ * data parameter.
+ * \param arch		architecture
+ * \param data		extra identification information (yasm_arch-specific)
+ *			[output]
+ * \param id		identifier as in the input file
+ * \param id_len	length of id string
+ * \param line		virtual line (from yasm_linemap)
+ * \return Identifier type (#YASM_ARCH_NOTREGTMOD if unrecognized)
+ */
+yasm_arch_regtmod yasm_arch_parse_check_regtmod
+    (yasm_arch *arch, /*@out@*/ unsigned long *data, const char *id,
+     size_t id_len, unsigned long line);
 
 /** Handle architecture-specific directives.
  * Should modify behavior ONLY of parse functions, much like parse_cpu().
@@ -384,54 +406,35 @@ int yasm_arch_parse_directive(yasm_arch *arch, const char *name,
 			      /*@null@*/ yasm_valparamhead *objext_valparams,
 			      yasm_object *object, unsigned long line);
 
-/** Create an instruction.  Creates a bytecode by matching the
- * instruction data and the parameters given with a valid instruction.
+/** Get NOP fill patterns for 1-15 bytes of fill.
  * \param arch		architecture
+ * \return 16-entry array of arrays; [0] is unused, [1] - [15] point to arrays
+ * of 1-15 bytes (respectively) in length.
+ */
+const unsigned char **yasm_arch_get_fill(const yasm_arch *arch);
+
+/** Finalize an instruction from a semi-generic insn description.  Note an
+ * existing bytecode is required.
+ * \param arch		architecture
+ * \param bc		bytecode to finalize
+ * \param prev_bc	previous bytecode in section
  * \param data		instruction data (from parse_check_id()); all
  *				zero indicates an empty instruction
- * \param num_operands	number of operands parsed
+ * \param num_operands	number of operands
  * \param operands	list of operands (in parse order)
- * \param prev_bc	previously parsed bytecode in section (NULL if
- *			first bytecode in section)
- * \param line		virtual line (from yasm_linemap)
- * \return If no match is found (the instruction is invalid), NULL,
- *	       otherwise newly allocated bytecode containing instruction.
+ * \param num_prefixes	number of prefixes
+ * \param prefixes	array of 4-element prefix data
+ * \param num_segregs	number of segment register prefixes
+ * \param segregs	array of segment register data
+ * \return If no match is found (the instruction is invalid), no action is
+ *         performed and an error is recorded.
  */
-/*@null@*/ yasm_bytecode *yasm_arch_parse_insn
-    (yasm_arch *arch, const unsigned long data[4], int num_operands,
-     /*@null@*/ yasm_insn_operands *operands, yasm_bytecode *prev_bc,
-     unsigned long line);
-
-/** Handle an instruction prefix.
- * Modifies an instruction bytecode based on the prefix in data.
- * \param arch	architecture
- * \param bc	bytecode (must be instruction bytecode)
- * \param data	prefix (from parse_check_id())
- * \param line	virtual line (from yasm_linemap)
- */
-void yasm_arch_parse_prefix(yasm_arch *arch, yasm_bytecode *bc,
-			    const unsigned long data[4], unsigned long line);
-
-/** Handle an segment register instruction prefix.
- * Modifies an instruction bytecode based on a segment register prefix.
- * \param arch		architecture
- * \param bc		bytecode (must be instruction bytecode)
- * \param segreg	segment register (from parse_check_id())
- * \param line		virtual line (from yasm_linemap)
- */
-void yasm_arch_parse_seg_prefix(yasm_arch *arch, yasm_bytecode *bc,
-				unsigned long segreg, unsigned long line);
-
-/** Handle a memory expression segment override.
- * Modifies an instruction bytecode based on a segment override in a
- * memory expression.
- * \param arch		architecture
- * \param ea		effective address
- * \param segreg	segment register (from parse_check_id())
- * \param line		virtual line (from yasm_linemap)
- */
-void yasm_arch_parse_seg_override(yasm_arch *arch, yasm_effaddr *ea,
-				  unsigned long segreg, unsigned long line);
+void yasm_arch_finalize_insn
+    (yasm_arch *arch, yasm_bytecode *bc, yasm_bytecode *prev_bc,
+     const unsigned long data[4], int num_operands,
+     /*@null@*/ yasm_insn_operands *operands, int num_prefixes,
+     const unsigned long **prefixes, int num_segregs,
+     const unsigned long *segregs);
 
 /** Output #yasm_floatnum to buffer.  Puts the value into the least
  * significant bits of the destination, or may be shifted into more
@@ -453,19 +456,6 @@ int yasm_arch_floatnum_tobytes(yasm_arch *arch, const yasm_floatnum *flt,
 			       size_t valsize, size_t shift, int warn,
 			       unsigned long line);
 
-/** Adjust #yasm_intnum for relative displacement from bc.  Displacement
- * is modified in-place.
- * \param arch		architecture
- * \param intn		integer value
- * \param valsize	size (in bits)
- * \param bc		bytecode being output ("parent" of value)
- * \param line		virtual line; may be 0 if warn is 0
- * \return Nonzero on error.
- */
-int yasm_arch_intnum_fixup_rel(yasm_arch *arch, yasm_intnum *intn,
-			       size_t valsize, const yasm_bytecode *bc,
-			       unsigned long line);
-
 /** Output #yasm_intnum to buffer.  Puts the value into the least
  * significant bits of the destination, or may be shifted into more
  * significant bits by the shift parameter.  The destination bits are
@@ -496,6 +486,18 @@ int yasm_arch_intnum_tobytes(yasm_arch *arch, const yasm_intnum *intn,
  */
 unsigned int yasm_arch_get_reg_size(yasm_arch *arch, unsigned long reg);
 
+/** Get a specific register of a register group, based on the register
+ * group and the index within the group.
+ * \param arch		architecture
+ * \param reggroup	register group
+ * \param regindex	register index
+ * \return 0 if regindex is not valid for that register group, otherwise the
+ *         specific register value.
+ */
+unsigned long yasm_arch_reggroup_get_reg(yasm_arch *arch,
+					 unsigned long reggroup,
+					 unsigned long regindex);
+
 /** Print a register.  For debugging purposes.
  * \param arch		architecture
  * \param reg		register
@@ -528,47 +530,52 @@ yasm_effaddr *yasm_arch_ea_create(yasm_arch *arch, /*@keep@*/ yasm_expr *e);
     (((yasm_arch_base *)arch)->module->keyword)
 #define yasm_arch_wordsize(arch) \
     (((yasm_arch_base *)arch)->module->wordsize)
+#define yasm_arch_min_insn_len(arch) \
+    (((yasm_arch_base *)arch)->module->min_insn_len)
 
-#define yasm_arch_create(module, machine)   module->create(machine)
+#define yasm_arch_create(module, machine, parser, error) \
+    module->create(machine, parser, error)
 
 #define yasm_arch_destroy(arch) \
     ((yasm_arch_base *)arch)->module->destroy(arch)
 #define yasm_arch_get_machine(arch) \
     ((yasm_arch_base *)arch)->module->get_machine(arch)
+#define yasm_arch_get_address_size(arch) \
+    ((yasm_arch_base *)arch)->module->get_address_size(arch)
 #define yasm_arch_set_var(arch, var, val) \
     ((yasm_arch_base *)arch)->module->set_var(arch, var, val)
-#define yasm_arch_parse_cpu(arch, cpuid, line) \
-    ((yasm_arch_base *)arch)->module->parse_cpu(arch, cpuid, line)
-#define yasm_arch_parse_check_id(arch, data, id, line) \
-    ((yasm_arch_base *)arch)->module->parse_check_id(arch, data, id, line)
+#define yasm_arch_parse_cpu(arch, cpuid, cpuid_len, line) \
+    ((yasm_arch_base *)arch)->module->parse_cpu(arch, cpuid, cpuid_len, line)
+#define yasm_arch_parse_check_insnprefix(arch, data, id, id_len, line) \
+    ((yasm_arch_base *)arch)->module->parse_check_insnprefix(arch, data, id, \
+							     id_len, line)
+#define yasm_arch_parse_check_regtmod(arch, data, id, id_len, line) \
+    ((yasm_arch_base *)arch)->module->parse_check_regtmod(arch, data, id, \
+							  id_len, line)
 #define yasm_arch_parse_directive(arch, name, valparams, objext_valparams, \
 				  object, line) \
     ((yasm_arch_base *)arch)->module->parse_directive \
 	(arch, name, valparams, objext_valparams, object, line)
-#define yasm_arch_parse_insn(arch, data, num_operands, operands, prev_bc, \
-			     line) \
-    ((yasm_arch_base *)arch)->module->parse_insn \
-	(arch, data, num_operands, operands, prev_bc, line)
-#define yasm_arch_parse_prefix(arch, bc, data, line) \
-    ((yasm_arch_base *)arch)->module->parse_prefix(arch, bc, data, line)
-#define yasm_arch_parse_seg_prefix(arch, bc, segreg, line) \
-    ((yasm_arch_base *)arch)->module->parse_seg_prefix(arch, bc, segreg, line)
-#define yasm_arch_parse_seg_override(arch, ea, segreg, line) \
-    ((yasm_arch_base *)arch)->module->parse_seg_override \
-	(arch, ea, segreg, line)
+#define yasm_arch_get_fill(arch) \
+    ((yasm_arch_base *)arch)->module->get_fill(arch)
+#define yasm_arch_finalize_insn(arch, bc, prev_bc, data, num_operands, \
+				operands, num_prefixes, prefixes, \
+				num_segregs, segregs) \
+    ((yasm_arch_base *)arch)->module->finalize_insn \
+	(arch, bc, prev_bc, data, num_operands, operands, num_prefixes, \
+	 prefixes, num_segregs, segregs)
 #define yasm_arch_floatnum_tobytes(arch, flt, buf, destsize, valsize, shift, \
 				   warn, line) \
     ((yasm_arch_base *)arch)->module->floatnum_tobytes \
 	(arch, flt, buf, destsize, valsize, shift, warn, line)
-#define yasm_arch_intnum_fixup_rel(arch, intn, valsize, bc, line) \
-    ((yasm_arch_base *)arch)->module->intnum_fixup_rel \
-	(arch, intn, valsize, bc, line)
 #define yasm_arch_intnum_tobytes(arch, intn, buf, destsize, valsize, shift, \
 				 bc, warn, line) \
     ((yasm_arch_base *)arch)->module->intnum_tobytes \
 	(arch, intn, buf, destsize, valsize, shift, bc, warn, line)
 #define yasm_arch_get_reg_size(arch, reg) \
     ((yasm_arch_base *)arch)->module->get_reg_size(arch, reg)
+#define yasm_arch_reggroup_get_reg(arch, regg, regi) \
+    ((yasm_arch_base *)arch)->module->reggroup_get_reg(arch, regg, regi)
 #define yasm_arch_reg_print(arch, reg, f) \
     ((yasm_arch_base *)arch)->module->reg_print(arch, reg, f)
 #define yasm_arch_segreg_print(arch, segreg, f) \
