@@ -26,7 +26,7 @@
  */
 #define YASM_LIB_INTERNAL
 #include "util.h"
-/*@unused@*/ RCSID("$Id: assocdat.c 1137 2004-09-04 01:24:57Z peter $");
+/*@unused@*/ RCSID("$Id: assocdat.c 1591 2006-07-13 03:25:38Z peter $");
 
 #include "coretype.h"
 #include "assocdat.h"
@@ -108,7 +108,7 @@ yasm__assoc_data_add(yasm__assoc_data *assoc_data_arg,
     }
 
     /* Delete existing data (if any) */
-    if (item->data)
+    if (item->data && item->data != data)
 	item->callback->destroy(item->data);
 
     item->data = data;
