@@ -26,7 +26,7 @@
  */
 #define YASM_LIB_INTERNAL
 #include "util.h"
-/*@unused@*/ RCSID("$Id: intnum.c 1158 2004-10-11 02:14:34Z peter $");
+/*@unused@*/ RCSID("$Id: intnum.c 1591 2006-07-13 03:25:38Z peter $");
 
 #include <ctype.h>
 #include <limits.h>
@@ -258,6 +258,44 @@ yasm_intnum_create_int(long i)
     return intn;
 }
 
+yasm_intnum *
+yasm_intnum_create_leb128(const unsigned char *ptr, int sign,
+			  unsigned long *size, unsigned long line)
+{
+    yasm_intnum *intn = yasm_xmalloc(sizeof(yasm_intnum));
+    const unsigned char *ptr_orig = ptr;
+    unsigned long i = 0;
+
+    intn->origsize = 0;
+
+    BitVector_Empty(conv_bv);
+    for (;;) {
+	BitVector_Chunk_Store(conv_bv, 7, i, *ptr);
+	i += 7;
+	if ((*ptr & 0x80) != 0x80)
+	    break;
+	ptr++;
+    }
+
+    *size = (ptr-ptr_orig)+1;
+
+    if(i > BITVECT_NATIVE_SIZE)
+	yasm__warning(YASM_WARN_GENERAL, line,
+		      N_("Numeric constant too large for internal format"));
+    else if (sign && (*ptr & 0x40) == 0x40)
+	BitVector_Interval_Fill(conv_bv, i, BITVECT_NATIVE_SIZE-1);
+
+    if (Set_Max(conv_bv) < 32) {
+	intn->type = INTNUM_UL;
+	intn->val.ul = BitVector_Chunk_Read(conv_bv, 32, 0);
+    } else {
+	intn->type = INTNUM_BV;
+	intn->val.bv = BitVector_Clone(conv_bv);
+    }
+
+    return intn;
+}
+
 yasm_intnum *
 yasm_intnum_copy(const yasm_intnum *intn)
 {
@@ -358,6 +396,10 @@ yasm_intnum_calc(yasm_intnum *acc, yasm_expr_op op, yasm_intnum *operand,
 	case YASM_EXPR_XOR:
 	    Set_ExclusiveOr(result, op1, op2);
 	    break;
+	case YASM_EXPR_NOR:
+	    Set_Union(result, op1, op2);
+	    Set_Complement(result, result);
+	    break;
 	case YASM_EXPR_SHL:
 	    if (operand->type == INTNUM_UL) {
 		BitVector_Copy(result, op1);
@@ -447,32 +489,50 @@ yasm_intnum_calc(yasm_intnum *acc, yasm_expr_op op, yasm_intnum *operand,
 
 void
 yasm_intnum_zero(yasm_intnum *intn)
+{
+    yasm_intnum_set_uint(intn, 0);
+}
+
+void
+yasm_intnum_set_uint(yasm_intnum *intn, unsigned long val)
 {
     if (intn->type == INTNUM_BV) {
 	BitVector_Destroy(intn->val.bv);
 	intn->type = INTNUM_UL;
     }
-    intn->val.ul = 0;
+    intn->val.ul = val;
 }
 
 int
-yasm_intnum_is_zero(yasm_intnum *intn)
+yasm_intnum_is_zero(const yasm_intnum *intn)
 {
     return (intn->type == INTNUM_UL && intn->val.ul == 0);
 }
 
 int
-yasm_intnum_is_pos1(yasm_intnum *intn)
+yasm_intnum_is_pos1(const yasm_intnum *intn)
 {
     return (intn->type == INTNUM_UL && intn->val.ul == 1);
 }
 
 int
-yasm_intnum_is_neg1(yasm_intnum *intn)
+yasm_intnum_is_neg1(const yasm_intnum *intn)
 {
     return (intn->type == INTNUM_BV && BitVector_is_full(intn->val.bv));
 }
 
+int
+yasm_intnum_sign(const yasm_intnum *intn)
+{
+    if (intn->type == INTNUM_UL) {
+	if (intn->val.ul == 0)
+	    return 0;
+	else
+	    return 1;
+    } else
+	return BitVector_Sign(intn->val.bv);
+}
+
 unsigned long
 yasm_intnum_get_uint(const yasm_intnum *intn)
 {
@@ -537,7 +597,11 @@ yasm_intnum_get_sized(const yasm_intnum *intn, unsigned char *ptr,
 	yasm_internal_error(N_("destination too large"));
 
     /* General size warnings */
-    if (warn && !yasm_intnum_check_size(intn, valsize, rshift, 2))
+    if (warn<0 && !yasm_intnum_check_size(intn, valsize, rshift, 1))
+	yasm__warning(YASM_WARN_GENERAL, line,
+		      N_("value does not fit in signed %d bit field"),
+		      valsize);
+    if (warn>0 && !yasm_intnum_check_size(intn, valsize, rshift, 2))
 	yasm__warning(YASM_WARN_GENERAL, line,
 		      N_("value does not fit in %d bit field"), valsize);
 
@@ -634,6 +698,167 @@ yasm_intnum_check_size(const yasm_intnum *intn, size_t size, size_t rshift,
     return (Set_Max(val) < (long)size);
 }
 
+static unsigned long
+get_leb128(wordptr val, unsigned char *ptr, int sign)
+{
+    unsigned long i, size;
+    unsigned char *ptr_orig = ptr;
+
+    if (sign) {
+	/* Signed mode */
+	if (BitVector_msb_(val)) {
+	    /* Negative */
+	    BitVector_Negate(conv_bv, val);
+	    size = Set_Max(conv_bv)+2;
+	} else {
+	    /* Positive */
+	    size = Set_Max(val)+2;
+	}
+    } else {
+	/* Unsigned mode */
+	size = Set_Max(val)+1;
+    }
+
+    /* Positive/Unsigned write */
+    for (i=0; i<size; i += 7) {
+	*ptr = (unsigned char)BitVector_Chunk_Read(val, 7, i);
+	*ptr |= 0x80;
+	ptr++;
+    }
+    *(ptr-1) &= 0x7F;	/* Clear MSB of last byte */
+    return (unsigned long)(ptr-ptr_orig);
+}
+
+static unsigned long
+size_leb128(wordptr val, int sign)
+{
+    if (sign) {
+	/* Signed mode */
+	if (BitVector_msb_(val)) {
+	    /* Negative */
+	    BitVector_Negate(conv_bv, val);
+	    return (Set_Max(conv_bv)+8)/7;
+	} else {
+	    /* Positive */
+	    return (Set_Max(val)+8)/7;
+	}
+    } else {
+	/* Unsigned mode */
+	return (Set_Max(val)+7)/7;
+    }
+}
+
+unsigned long
+yasm_intnum_get_leb128(const yasm_intnum *intn, unsigned char *ptr, int sign)
+{
+    wordptr val = op1static;
+
+    /* Shortcut 0 */
+    if (intn->type == INTNUM_UL && intn->val.ul == 0) {
+	*ptr = 0;
+	return 1;
+    }
+
+    /* If not already a bitvect, convert value to be written to a bitvect */
+    if (intn->type == INTNUM_BV)
+	val = intn->val.bv;
+    else {
+	BitVector_Empty(val);
+	BitVector_Chunk_Store(val, 32, 0, intn->val.ul);
+    }
+
+    return get_leb128(val, ptr, sign);
+}
+
+unsigned long
+yasm_intnum_size_leb128(const yasm_intnum *intn, int sign)
+{
+    wordptr val = op1static;
+
+    /* Shortcut 0 */
+    if (intn->type == INTNUM_UL && intn->val.ul == 0) {
+	return 1;
+    }
+
+    /* If not already a bitvect, convert value to a bitvect */
+    if (intn->type == INTNUM_BV)
+	val = intn->val.bv;
+    else {
+	BitVector_Empty(val);
+	BitVector_Chunk_Store(val, 32, 0, intn->val.ul);
+    }
+
+    return size_leb128(val, sign);
+}
+
+unsigned long
+yasm_get_sleb128(long v, unsigned char *ptr)
+{
+    wordptr val = op1static;
+
+    /* Shortcut 0 */
+    if (v == 0) {
+	*ptr = 0;
+	return 1;
+    }
+
+    BitVector_Empty(val);
+    if (v >= 0)
+	BitVector_Chunk_Store(val, 32, 0, (unsigned long)v);
+    else {
+	BitVector_Chunk_Store(val, 32, 0, (unsigned long)(-v));
+	BitVector_Negate(val, val);
+    }
+    return get_leb128(val, ptr, 1);
+}
+
+unsigned long
+yasm_size_sleb128(long v)
+{
+    wordptr val = op1static;
+
+    if (v == 0)
+	return 1;
+
+    BitVector_Empty(val);
+    if (v >= 0)
+	BitVector_Chunk_Store(val, 32, 0, (unsigned long)v);
+    else {
+	BitVector_Chunk_Store(val, 32, 0, (unsigned long)(-v));
+	BitVector_Negate(val, val);
+    }
+    return size_leb128(val, 1);
+}
+
+unsigned long
+yasm_get_uleb128(unsigned long v, unsigned char *ptr)
+{
+    wordptr val = op1static;
+
+    /* Shortcut 0 */
+    if (v == 0) {
+	*ptr = 0;
+	return 1;
+    }
+
+    BitVector_Empty(val);
+    BitVector_Chunk_Store(val, 32, 0, v);
+    return get_leb128(val, ptr, 0);
+}
+
+unsigned long
+yasm_size_uleb128(unsigned long v)
+{
+    wordptr val = op1static;
+
+    if (v == 0)
+	return 1;
+
+    BitVector_Empty(val);
+    BitVector_Chunk_Store(val, 32, 0, v);
+    return size_leb128(val, 0);
+}
+
 void
 yasm_intnum_print(const yasm_intnum *intn, FILE *f)
 {
