@@ -26,7 +26,7 @@
  */
 #define YASM_LIB_INTERNAL
 #include "util.h"
-/*@unused@*/ RCSID("$Id: symrec.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: symrec.c 1591 2006-07-13 03:25:38Z peter $");
 
 #ifdef STDC_HEADERS
 # include <limits.h>
@@ -58,7 +58,12 @@ typedef enum {
 typedef enum {
     SYM_UNKNOWN,		/* for unknown type (COMMON/EXTERN) */
     SYM_EQU,			/* for EQU defined symbols (expressions) */
-    SYM_LABEL			/* for labels */
+    SYM_LABEL,			/* for labels */
+    SYM_CURPOS,			/* for labels representing the current
+				   assembly position */
+    SYM_SPECIAL			/* for special symbols that need to be in
+				   the symbol table but otherwise have no
+				   purpose */
 } sym_type;
 
 struct yasm_symrec {
@@ -163,8 +168,6 @@ symtab_get_or_new(yasm_symtab *symtab, const char *name, int in_table)
 }
 /*@=freshtrans =mustfree@*/
 
-/* Call a function with each symrec.  Stops early if 0 returned by func.
-   Returns 0 if stopped early. */
 int
 yasm_symtab_traverse(yasm_symtab *symtab, void *d,
 		     int (*func) (yasm_symrec *sym, void *d))
@@ -182,6 +185,12 @@ yasm_symtab_use(yasm_symtab *symtab, const char *name, unsigned long line)
     return rec;
 }
 
+yasm_symrec *
+yasm_symtab_get(yasm_symtab *symtab, const char *name)
+{
+    return HAMT_search(symtab->sym_table, name);
+}
+
 static /*@dependent@*/ yasm_symrec *
 symtab_define(yasm_symtab *symtab, const char *name, sym_type type,
 	      int in_table, unsigned long line)
@@ -189,12 +198,14 @@ symtab_define(yasm_symtab *symtab, const char *name, sym_type type,
     yasm_symrec *rec = symtab_get_or_new(symtab, name, in_table);
 
     /* Has it been defined before (either by DEFINED or COMMON/EXTERN)? */
-    if ((rec->status & SYM_DEFINED) ||
-	(rec->visibility & (YASM_SYM_COMMON | YASM_SYM_EXTERN))) {
+    if (rec->status & SYM_DEFINED) {
 	yasm__error(line, N_("redefinition of `%s'"), name);
 	yasm__error_at(line, rec->line, N_("`%s' previously defined here"),
 		       name);
     } else {
+	if (rec->visibility & YASM_SYM_EXTERN)
+	    yasm__warning(YASM_WARN_GENERAL, line,
+			  N_("`%s' both defined and declared extern"), name);
 	rec->line = line;	/* set line number of definition */
 	rec->type = type;
 	rec->status |= SYM_DEFINED;
@@ -226,12 +237,23 @@ yasm_symtab_define_label(yasm_symtab *symtab, const char *name,
 }
 
 yasm_symrec *
-yasm_symtab_define_label2(const char *name, yasm_bytecode *precbc,
-			  int in_table, unsigned long line)
+yasm_symtab_define_curpos(yasm_symtab *symtab, const char *name,
+			  yasm_bytecode *precbc, unsigned long line)
 {
-    return yasm_symtab_define_label(yasm_object_get_symtab(
-	yasm_section_get_object(yasm_bc_get_section(precbc))), name, precbc,
-	in_table, line);
+    yasm_symrec *rec;
+    rec = symtab_define(symtab, name, SYM_CURPOS, 0, line);
+    rec->value.precbc = precbc;
+    return rec;
+}
+
+yasm_symrec *
+yasm_symtab_define_special(yasm_symtab *symtab, const char *name,
+			   yasm_sym_vis vis)
+{
+    yasm_symrec *rec = symtab_define(symtab, name, SYM_SPECIAL, 1, 0);
+    rec->status |= SYM_VALUED;
+    rec->visibility = vis;
+    return rec;
 }
 
 yasm_symrec *
@@ -239,7 +261,13 @@ yasm_symtab_declare(yasm_symtab *symtab, const char *name, yasm_sym_vis vis,
 		    unsigned long line)
 {
     yasm_symrec *rec = symtab_get_or_new(symtab, name, 1);
+    yasm_symrec_declare(rec, vis, line);
+    return rec;
+}
 
+void
+yasm_symrec_declare(yasm_symrec *rec, yasm_sym_vis vis, unsigned long line)
+{
     /* Allowable combinations:
      *  Existing State--------------  vis  New State-------------------
      *  DEFINED GLOBAL COMMON EXTERN  GCE  DEFINED GLOBAL COMMON EXTERN
@@ -260,34 +288,46 @@ yasm_symtab_declare(yasm_symtab *symtab, const char *name, yasm_sym_vis vis,
     else
 	yasm__error(line,
 	    N_("duplicate definition of `%s'; first defined on line %lu"),
-	    name, rec->line);
-    return rec;
+	    rec->name, rec->line);
 }
 
+typedef struct symtab_finalize_info {
+    unsigned long firstundef_line;
+    int undef_extern;
+    /*@null@*/ yasm_objfmt *objfmt;
+} symtab_finalize_info;
+
 static int
 symtab_parser_finalize_checksym(yasm_symrec *sym, /*@null@*/ void *d)
 {
-    unsigned long *firstundef_line = d;
+    symtab_finalize_info *info = (symtab_finalize_info *)d;
     /* error if a symbol is used but never defined or extern/common declared */
     if ((sym->status & SYM_USED) && !(sym->status & SYM_DEFINED) &&
 	!(sym->visibility & (YASM_SYM_EXTERN | YASM_SYM_COMMON))) {
-	yasm__error(sym->line, N_("undefined symbol `%s' (first use)"),
-		    sym->name);
-	if (sym->line < *firstundef_line)
-	    *firstundef_line = sym->line;
+	if (info->undef_extern && info->objfmt)
+	    yasm_objfmt_extern_declare(info->objfmt, sym->name, NULL, 1);
+	else {
+	    yasm__error(sym->line, N_("undefined symbol `%s' (first use)"),
+			sym->name);
+	    if (sym->line < info->firstundef_line)
+		info->firstundef_line = sym->line;
+	}
     }
 
-    return 1;
+    return 0;
 }
 
 void
-yasm_symtab_parser_finalize(yasm_symtab *symtab)
+yasm_symtab_parser_finalize(yasm_symtab *symtab, int undef_extern,
+			    yasm_objfmt *objfmt)
 {
-    unsigned long firstundef_line = ULONG_MAX;
-    yasm_symtab_traverse(symtab, &firstundef_line,
-			 symtab_parser_finalize_checksym);
-    if (firstundef_line < ULONG_MAX)
-	yasm__error(firstundef_line,
+    symtab_finalize_info info;
+    info.firstundef_line = ULONG_MAX;
+    info.undef_extern = undef_extern;
+    info.objfmt = objfmt;
+    yasm_symtab_traverse(symtab, &info, symtab_parser_finalize_checksym);
+    if (info.firstundef_line < ULONG_MAX)
+	yasm__error(info.firstundef_line,
 		    N_(" (Each undefined symbol is reported only once.)"));
 }
 
@@ -319,7 +359,7 @@ symrec_print_wrapper(yasm_symrec *sym, /*@null@*/ void *d)
     assert(data != NULL);
     fprintf(data->f, "%*sSymbol `%s'\n", data->indent_level, "", sym->name);
     yasm_symrec_print(sym, data->f, data->indent_level+1);
-    return 1;
+    return 0;
 }
 
 void
@@ -356,7 +396,8 @@ int
 yasm_symrec_get_label(const yasm_symrec *sym,
 		      yasm_symrec_get_label_bytecodep *precbc)
 {
-    if (sym->type != SYM_LABEL || !sym->value.precbc) {
+    if (!(sym->type == SYM_LABEL || sym->type == SYM_CURPOS)
+	|| !sym->value.precbc) {
 	*precbc = (yasm_symrec_get_label_bytecodep)0xDEADBEEF;
 	return 0;
     }
@@ -364,6 +405,18 @@ yasm_symrec_get_label(const yasm_symrec *sym,
     return 1;
 }
 
+int
+yasm_symrec_is_special(const yasm_symrec *sym)
+{
+    return (sym->type == SYM_SPECIAL);
+}
+
+int
+yasm_symrec_is_curpos(const yasm_symrec *sym)
+{
+    return (sym->type == SYM_CURPOS);
+}
+
 void *
 yasm_symrec_get_data(yasm_symrec *sym,
 		     const yasm_assoc_data_callback *callback)
@@ -392,13 +445,18 @@ yasm_symrec_print(const yasm_symrec *sym, FILE *f, int indent_level)
 	    fprintf(f, "\n");
 	    break;
 	case SYM_LABEL:
-	    fprintf(f, "%*s_Label_\n%*sSection:\n", indent_level, "",
+	case SYM_CURPOS:
+	    fprintf(f, "%*s_%s_\n%*sSection:\n", indent_level, "",
+		    sym->type == SYM_LABEL ? "Label" : "CurPos",
 		    indent_level, "");
 	    yasm_section_print(yasm_bc_get_section(sym->value.precbc), f,
 			       indent_level+1, 0);
 	    fprintf(f, "%*sPreceding bytecode:\n", indent_level, "");
 	    yasm_bc_print(sym->value.precbc, f, indent_level+1);
 	    break;
+	case SYM_SPECIAL:
+	    fprintf(f, "%*s-Special-\n", indent_level, "");
+	    break;
     }
 
     fprintf(f, "%*sStatus=", indent_level, "");
