@@ -30,7 +30,7 @@
  */
 #define YASM_LIB_INTERNAL
 #include "util.h"
-RCSID("$Id: xstrdup.c 1137 2004-09-04 01:24:57Z peter $");
+RCSID("$Id: xstrdup.c 1591 2006-07-13 03:25:38Z peter $");
 
 #include "coretype.h"
 
@@ -39,16 +39,6 @@ static char sccsid[] = "@(#)strdup.c	8.1 (Berkeley) 6/4/93";
 #endif /* LIBC_SCCS and not lint */
 
 
-#ifndef STDC_HEADERS
-size_t strlen(const char *);
-# ifndef HAVE_MEMCPY
-void bcopy(const void *, void *, size_t);
-#  define memcpy(d, s, n) bcopy((s), (d), (n))
-# else
-void memcpy(void *, const void *, size_t);
-# endif
-#endif
-
 #ifdef WITH_DMALLOC
 #undef yasm__xstrdup
 #endif
