@@ -25,7 +25,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: x86arch.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: x86arch.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #include <libyasm.h>
@@ -37,17 +37,22 @@ yasm_arch_module yasm_x86_LTX_arch;
 
 
 static /*@only@*/ yasm_arch *
-x86_create(const char *machine)
+x86_create(const char *machine, const char *parser,
+	   /*@out@*/ yasm_arch_create_error *error)
 {
     yasm_arch_x86 *arch_x86;
     unsigned int amd64_machine;
 
+    *error = YASM_ARCH_CREATE_OK;
+
     if (yasm__strcasecmp(machine, "x86") == 0)
 	amd64_machine = 0;
     else if (yasm__strcasecmp(machine, "amd64") == 0)
 	amd64_machine = 1;
-    else
+    else {
+	*error = YASM_ARCH_CREATE_BAD_MACHINE;
 	return NULL;
+    }
 
     arch_x86 = yasm_xmalloc(sizeof(yasm_arch_x86));
 
@@ -57,6 +62,16 @@ x86_create(const char *machine)
     arch_x86->amd64_machine = amd64_machine;
     arch_x86->mode_bits = 0;
 
+    if (yasm__strcasecmp(parser, "nasm") == 0)
+	arch_x86->parser = X86_PARSER_NASM;
+    else if (yasm__strcasecmp(parser, "gas") == 0)
+	arch_x86->parser = X86_PARSER_GAS;
+    else {
+	yasm_xfree(arch_x86);
+	*error = YASM_ARCH_CREATE_BAD_PARSER;
+	return NULL;
+    }
+
     return (yasm_arch *)arch_x86;
 }
 
@@ -76,6 +91,18 @@ x86_get_machine(const yasm_arch *arch)
 	return "x86";
 }
 
+static unsigned int
+x86_get_address_size(const yasm_arch *arch)
+{
+    const yasm_arch_x86 *arch_x86 = (const yasm_arch_x86 *)arch;
+    if (arch_x86->mode_bits != 0)
+	return arch_x86->mode_bits;
+    if (arch_x86->amd64_machine)
+	return 64;
+    else
+	return 32;
+}
+
 static int
 x86_set_var(yasm_arch *arch, const char *var, unsigned long val)
 {
@@ -113,6 +140,125 @@ x86_parse_directive(yasm_arch *arch, const char *name,
 	return 1;
 }
 
+static const unsigned char **
+x86_get_fill(const yasm_arch *arch)
+{
+    const yasm_arch_x86 *arch_x86 = (const yasm_arch_x86 *)arch;
+
+    /* Fill patterns that GAS uses. */
+    static const char *fill16[16] = {
+	NULL,				/* unused			*/
+	"\x90",				/* 1 - nop			*/
+	"\x89\xf6",			/* 2 - mov si, si		*/
+	"\x8d\x74\x00",			/* 3 - lea si, [si+byte 0]	*/
+	"\x8d\xb4\x00\x00",		/* 4 - lea si, [si+word 0]	*/
+	"\x90"				/* 5 - nop			*/
+	"\x8d\xb4\x00\x00",		/*     lea si, [si+word 0]	*/
+	"\x89\xf6"			/* 6 - mov si, si		*/
+	"\x8d\xbd\x00\x00",		/*     lea di, [di+word 0]	*/
+	"\x8d\x74\x00"			/* 7 - lea si, [si+byte 0]	*/
+	"\x8d\xbd\x00\x00",		/*     lea di, [di+word 0]	*/
+	"\x8d\xb4\x00\x00"		/* 8 - lea si, [si+word 0]	*/
+	"\x8d\xbd\x00\x00",		/*     lea di, [di+word 0]	*/
+	"\xeb\x07\x90\x90\x90\x90\x90"	/* 9 - jmp $+9; nop fill	*/
+	"\x90\x90",
+	"\xeb\x08\x90\x90\x90\x90\x90"	/* 10 - jmp $+10; nop fill	*/
+	"\x90\x90\x90",
+	"\xeb\x09\x90\x90\x90\x90\x90"	/* 11 - jmp $+11; nop fill	*/
+	"\x90\x90\x90\x90",
+	"\xeb\x0a\x90\x90\x90\x90\x90"	/* 12 - jmp $+12; nop fill	*/
+	"\x90\x90\x90\x90\x90",
+	"\xeb\x0b\x90\x90\x90\x90\x90"	/* 13 - jmp $+13; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90",
+	"\xeb\x0c\x90\x90\x90\x90\x90"	/* 14 - jmp $+14; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90\x90",
+	"\xeb\x0d\x90\x90\x90\x90\x90"	/* 15 - jmp $+15; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90\x90\x90"
+    };
+    static const char *fill32[16] = {
+	NULL,				/* unused			*/
+	"\x90",				/* 1 - nop			*/
+	"\x89\xf6",			/* 2 - mov esi, esi		*/
+	"\x8d\x76\x00",			/* 3 - lea esi, [esi+byte 0]	*/
+	"\x8d\x74\x26\x00",		/* 4 - lea esi, [esi*1+byte 0]	*/
+	"\x90"				/* 5 - nop			*/
+	"\x8d\x74\x26\x00",		/*     lea esi, [esi*1+byte 0]	*/
+	"\x8d\xb6\x00\x00\x00\x00",	/* 6 - lea esi, [esi+dword 0]	*/
+	"\x8d\xb4\x26\x00\x00\x00\x00",	/* 7 - lea esi, [esi*1+dword 0]	*/
+	"\x90"				/* 8 - nop			*/
+	"\x8d\xb4\x26\x00\x00\x00\x00",	/*     lea esi, [esi*1+dword 0]	*/
+#if 0
+	/* GAS uses these */
+	"\x89\xf6"			/* 9 - mov esi, esi		*/
+	"\x8d\xbc\x27\x00\x00\x00\x00",	/*     lea edi, [edi*1+dword 0]	*/
+	"\x8d\x76\x00"			/* 10 - lea esi, [esi+byte 0]	*/
+	"\x8d\xbc\x27\x00\x00\x00\x00",	/*      lea edi, [edi+dword 0]	*/
+	"\x8d\x74\x26\x00"		/* 11 - lea esi, [esi*1+byte 0]	*/
+	"\x8d\xbc\x27\x00\x00\x00\x00",	/*      lea edi, [edi*1+dword 0]*/
+	"\x8d\xb6\x00\x00\x00\x00"	/* 12 - lea esi, [esi+dword 0]	*/
+	"\x8d\xbf\x00\x00\x00\x00",	/*      lea edi, [edi+dword 0]	*/
+	"\x8d\xb6\x00\x00\x00\x00"	/* 13 - lea esi, [esi+dword 0]	*/
+	"\x8d\xbc\x27\x00\x00\x00\x00",	/*      lea edi, [edi*1+dword 0]*/
+	"\x8d\xb4\x26\x00\x00\x00\x00"	/* 14 - lea esi, [esi*1+dword 0]*/
+	"\x8d\xbc\x27\x00\x00\x00\x00",	/*      lea edi, [edi*1+dword 0]*/
+#else
+	/* But on newer processors, these are recommended */
+	"\xeb\x07\x90\x90\x90\x90\x90"	/* 9 - jmp $+9; nop fill	*/
+	"\x90\x90",
+	"\xeb\x08\x90\x90\x90\x90\x90"	/* 10 - jmp $+10; nop fill	*/
+	"\x90\x90\x90",
+	"\xeb\x09\x90\x90\x90\x90\x90"	/* 11 - jmp $+11; nop fill	*/
+	"\x90\x90\x90\x90",
+	"\xeb\x0a\x90\x90\x90\x90\x90"	/* 12 - jmp $+12; nop fill	*/
+	"\x90\x90\x90\x90\x90",
+	"\xeb\x0b\x90\x90\x90\x90\x90"	/* 13 - jmp $+13; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90",
+	"\xeb\x0c\x90\x90\x90\x90\x90"	/* 14 - jmp $+14; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90\x90",
+#endif
+	"\xeb\x0d\x90\x90\x90\x90\x90"	/* 15 - jmp $+15; nop fill	*/
+	"\x90\x90\x90\x90\x90\x90\x90\x90"
+    };
+    static const char *fill64[16] = {
+	NULL,				/* unused			*/
+	"\x90",				/* 1 - nop			*/
+	"\x66\x90",			/* 2 - o16; nop			*/
+	"\x66\x66\x90",			/* 3 - o16; o16; nop		*/
+	"\x66\x66\x66\x90",		/* 4 - o16; o16; o16; nop	*/
+	"\x66\x66\x90\x66\x90",		/* 5 */
+	"\x66\x66\x90\x66\x66\x90",	/* 6 */
+	"\x66\x66\x66\x90\x66\x66\x90",	/* 7 */
+	"\x66\x66\x66\x90\x66\x66\x66"	/* 8 */
+	"\x90",
+	"\x66\x66\x90\x66\x66\x90\x66"	/* 9 */
+	"\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x90"	/* 10 */
+	"\x66\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x66"	/* 11 */
+	"\x90\x66\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x66"	/* 12 */
+	"\x90\x66\x66\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x90"	/* 13 */
+	"\x66\x66\x90\x66\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x66"	/* 14 */
+	"\x90\x66\x66\x90\x66\x66\x90",
+	"\x66\x66\x66\x90\x66\x66\x66"	/* 15 */
+	"\x90\x66\x66\x66\x90\x66\x66\x90"
+    };
+    switch (arch_x86->mode_bits) {
+	case 16:
+	    return (const unsigned char **)fill16;
+	case 32:
+	    return (const unsigned char **)fill32;
+	case 64:
+	    return (const unsigned char **)fill64;
+	default:
+	    yasm_internal_error(N_("Invalid mode_bits in x86_get_fill"));
+	    /*@notreached@*/
+	    return NULL;
+    }
+}
+
 unsigned int
 yasm_x86__get_reg_size(yasm_arch *arch, unsigned long reg)
 {
@@ -140,6 +286,30 @@ yasm_x86__get_reg_size(yasm_arch *arch, unsigned long reg)
     return 0;
 }
 
+static unsigned long
+x86_reggroup_get_reg(yasm_arch *arch, unsigned long reggroup,
+		     unsigned long regindex)
+{
+    yasm_arch_x86 *arch_x86 = (yasm_arch_x86 *)arch;
+    switch ((x86_expritem_reg_size)(reggroup & ~0xFUL)) {
+	case X86_XMMREG:
+	    if (arch_x86->mode_bits == 64) {
+		if (regindex > 15)
+		    return 0;
+		return reggroup | (regindex & 15);
+	    }
+	    /*@fallthrough@*/
+	case X86_MMXREG:
+	case X86_FPUREG:
+	    if (regindex > 7)
+		return 0;
+	    return reggroup | (regindex & 7);
+	default:
+	    yasm_internal_error(N_("bad register group"));
+    }
+    return 0;
+}
+
 static void
 x86_reg_print(yasm_arch *arch, unsigned long reg, FILE *f)
 {
@@ -207,38 +377,6 @@ x86_segreg_print(yasm_arch *arch, unsigned long segreg, FILE *f)
     fprintf(f, "%s", name[segreg&7]);
 }
 
-static void
-x86_parse_prefix(yasm_arch *arch, yasm_bytecode *bc,
-		 const unsigned long data[4], unsigned long line)
-{
-    switch((x86_parse_insn_prefix)data[0]) {
-	case X86_LOCKREP:
-	    yasm_x86__bc_insn_set_lockrep_prefix(bc, data[1] & 0xff, line);
-	    break;
-	case X86_ADDRSIZE:
-	    yasm_x86__bc_insn_addrsize_override(bc, data[1]);
-	    break;
-	case X86_OPERSIZE:
-	    yasm_x86__bc_insn_opersize_override(bc, data[1]);
-	    break;
-    }
-}
-
-static void
-x86_parse_seg_prefix(yasm_arch *arch, yasm_bytecode *bc, unsigned long segreg,
-		     unsigned long line)
-{
-    yasm_x86__ea_set_segment(yasm_x86__bc_insn_get_ea(bc),
-			     (unsigned char)(segreg>>8), line);
-}
-
-static void
-x86_parse_seg_override(yasm_arch *arch, yasm_effaddr *ea,
-		       unsigned long segreg, unsigned long line)
-{
-    yasm_x86__ea_set_segment(ea, (unsigned char)(segreg>>8), line);
-}
-
 /* Define x86 machines -- see arch.h for details */
 static yasm_arch_machine x86_machines[] = {
     { "IA-32 and derivatives", "x86" },
@@ -248,28 +386,28 @@ static yasm_arch_machine x86_machines[] = {
 
 /* Define arch structure -- see arch.h for details */
 yasm_arch_module yasm_x86_LTX_arch = {
-    YASM_ARCH_VERSION,
     "x86 (IA-32 and derivatives), AMD64",
     "x86",
     x86_create,
     x86_destroy,
     x86_get_machine,
+    x86_get_address_size,
     x86_set_var,
     yasm_x86__parse_cpu,
-    yasm_x86__parse_check_id,
+    yasm_x86__parse_check_insnprefix,
+    yasm_x86__parse_check_regtmod,
     x86_parse_directive,
-    yasm_x86__parse_insn,
-    x86_parse_prefix,
-    x86_parse_seg_prefix,
-    x86_parse_seg_override,
+    x86_get_fill,
+    yasm_x86__finalize_insn,
     yasm_x86__floatnum_tobytes,
-    yasm_x86__intnum_fixup_rel,
     yasm_x86__intnum_tobytes,
     yasm_x86__get_reg_size,
+    x86_reggroup_get_reg,
     x86_reg_print,
     x86_segreg_print,
     yasm_x86__ea_create_expr,
     x86_machines,
     "x86",
-    2
+    2,
+    1
 };
