@@ -25,7 +25,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: stabs-dbgfmt.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: stabs-dbgfmt.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_BC_INTERNAL
@@ -85,7 +85,6 @@ typedef struct yasm_dbgfmt_stabs {
 
     yasm_object *object;
     yasm_symtab *symtab;
-    const char *filename;
     yasm_linemap *linemap;
     yasm_arch *arch;
 } yasm_dbgfmt_stabs;
@@ -117,38 +116,26 @@ typedef struct {
     unsigned long value;		/* fallthrough value if above NULL */
 } stabs_stab;
 
-/* Bytecode types */
-
-typedef struct {
-    yasm_bytecode bc;	/* base structure */
-    /*@only@*/ char *str;
-} stabs_bc_str;
-
-typedef struct {
-    yasm_bytecode bc;	/* base structure */
-    stabs_stab *stab;
-} stabs_bc_stab;
-
 /* Bytecode callback function prototypes */
 
-static void stabs_bc_str_destroy(yasm_bytecode *bc);
-static void stabs_bc_str_print(const yasm_bytecode *bc, FILE *f, int
+static void stabs_bc_str_destroy(void *contents);
+static void stabs_bc_str_print(const void *contents, FILE *f, int
 			       indent_level);
 static yasm_bc_resolve_flags stabs_bc_str_resolve
     (yasm_bytecode *bc, int save, yasm_calc_bc_dist_func calc_bc_dist);
 static int stabs_bc_str_tobytes
     (yasm_bytecode *bc, unsigned char **bufp, void *d,
-     yasm_output_expr_func output_expr,
+     yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
-static void stabs_bc_stab_destroy(yasm_bytecode *bc);
-static void stabs_bc_stab_print(const yasm_bytecode *bc, FILE *f, int
+static void stabs_bc_stab_destroy(void *contents);
+static void stabs_bc_stab_print(const void *contents, FILE *f, int
 				indent_level);
 static yasm_bc_resolve_flags stabs_bc_stab_resolve
     (yasm_bytecode *bc, int save, yasm_calc_bc_dist_func calc_bc_dist);
 static int stabs_bc_stab_tobytes
     (yasm_bytecode *bc, unsigned char **bufp, void *d,
-     yasm_output_expr_func output_expr,
+     yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
 /* Bytecode callback structures */
@@ -156,6 +143,7 @@ static int stabs_bc_stab_tobytes
 static const yasm_bytecode_callback stabs_bc_str_callback = {
     stabs_bc_str_destroy,
     stabs_bc_str_print,
+    yasm_bc_finalize_common,
     stabs_bc_str_resolve,
     stabs_bc_str_tobytes
 };
@@ -163,6 +151,7 @@ static const yasm_bytecode_callback stabs_bc_str_callback = {
 static const yasm_bytecode_callback stabs_bc_stab_callback = {
     stabs_bc_stab_destroy,
     stabs_bc_stab_print,
+    yasm_bc_finalize_common,
     stabs_bc_stab_resolve,
     stabs_bc_stab_tobytes
 };
@@ -171,12 +160,10 @@ yasm_dbgfmt_module yasm_stabs_LTX_dbgfmt;
 
 
 static /*@null@*/ /*@only@*/ yasm_dbgfmt *
-stabs_dbgfmt_create(const char *in_filename, const char *obj_filename,
-		    yasm_object *object, yasm_objfmt *of, yasm_arch *a)
+stabs_dbgfmt_create(yasm_object *object, yasm_objfmt *of, yasm_arch *a)
 {
     yasm_dbgfmt_stabs *dbgfmt_stabs = yasm_xmalloc(sizeof(yasm_dbgfmt_stabs));
     dbgfmt_stabs->dbgfmt.module = &yasm_stabs_LTX_dbgfmt;
-    dbgfmt_stabs->filename = in_filename;
     dbgfmt_stabs->object = object;
     dbgfmt_stabs->symtab = yasm_object_get_symtab(object);
     dbgfmt_stabs->linemap = yasm_object_get_linemap(object);
@@ -197,15 +184,9 @@ static yasm_bytecode *
 stabs_dbgfmt_append_bcstr(yasm_section *sect, const char *str)
 {
     yasm_bytecode *bc, *precbc;
-    stabs_bc_str *bc_str;
    
     precbc = yasm_section_bcs_last(sect);
-    bc = yasm_bc_create_common(&stabs_bc_str_callback, sizeof(stabs_bc_str),
-			       0);
-    bc_str = (stabs_bc_str *)bc;
-
-    bc_str->str = yasm__xstrdup(str);
-
+    bc = yasm_bc_create_common(&stabs_bc_str_callback, yasm__xstrdup(str), 0);
     bc->len = strlen(str)+1;
     bc->offset = precbc ? precbc->offset + precbc->len : 0;
 
@@ -224,7 +205,6 @@ stabs_dbgfmt_append_stab(stabs_info *info, yasm_section *sect,
 			 /*@null@*/ yasm_bytecode *bcvalue, unsigned long value)
 {
     yasm_bytecode *bc, *precbc;
-    stabs_bc_stab *bc_stab;
     stabs_stab *stab = yasm_xmalloc(sizeof(stabs_stab));
 
     stab->other = 0;
@@ -236,12 +216,8 @@ stabs_dbgfmt_append_stab(stabs_info *info, yasm_section *sect,
     stab->value = value;
 
     precbc = yasm_section_bcs_last(sect);
-    bc = yasm_bc_create_common(&stabs_bc_stab_callback, sizeof(stabs_bc_stab),
+    bc = yasm_bc_create_common(&stabs_bc_stab_callback, stab,
 			       bcvalue ? bcvalue->line : 0);
-    bc_stab = (stabs_bc_stab *)bc;
-
-    bc_stab->stab = stab;
-
     bc->len = info->stablen;
     bc->offset = precbc ? precbc->offset + precbc->len : 0;
 
@@ -328,7 +304,9 @@ stabs_dbgfmt_generate_sections(yasm_section *sect, /*@null@*/ void *d)
     if (yasm__strcasecmp(sectname, ".text")==0) {
         /* Close out last function by appending a null SO stab after last bc */
         yasm_bytecode *bc = yasm_section_bcs_last(sect);
-        yasm_symrec *sym = yasm_symtab_define_label2(".n_so", bc, 1, bc->line);
+        yasm_symrec *sym =
+	    yasm_symtab_define_label(yasm_object_get_symtab(
+		yasm_section_get_object(sect)), ".n_so", bc, 1, bc->line);
 	stabs_dbgfmt_append_stab(info, info->stab, 0, N_SO, 0, sym, bc, 0);
     }
 
@@ -342,7 +320,6 @@ stabs_dbgfmt_generate(yasm_dbgfmt *dbgfmt)
     stabs_info info;
     int new;
     yasm_bytecode *dbgbc;
-    stabs_bc_stab *dbgbc_stab;
     stabs_stab *stab;
     yasm_bytecode *filebc, *nullbc, *laststr, *firstbc;
     yasm_symrec *firstsym;
@@ -358,8 +335,8 @@ stabs_dbgfmt_generate(yasm_dbgfmt *dbgfmt)
     info.dbgfmt_stabs = dbgfmt_stabs;
     info.lastline = 0;
     info.stabcount = 0;
-    info.stab = yasm_object_get_general(dbgfmt_stabs->object, ".stab", 0, 0,
-					&new, 0);
+    info.stab = yasm_object_get_general(dbgfmt_stabs->object, ".stab", 0, 4, 0,
+					0, &new, 0);
     if (!new) {
 	yasm_bytecode *last = yasm_section_bcs_last(info.stab);
 	if (last == NULL)
@@ -371,7 +348,7 @@ stabs_dbgfmt_generate(yasm_dbgfmt *dbgfmt)
     }
 
     info.stabstr = yasm_object_get_general(dbgfmt_stabs->object, ".stabstr", 0,
-					   0, &new, 0);
+					   1, 0, 0, &new, 0);
     if (!new) {
 	yasm_bytecode *last = yasm_section_bcs_last(info.stabstr);
 	if (last == NULL)
@@ -386,18 +363,14 @@ stabs_dbgfmt_generate(yasm_dbgfmt *dbgfmt)
 
     /* initial pseudo-stab */
     stab = yasm_xmalloc(sizeof(stabs_stab));
-    dbgbc = yasm_bc_create_common(&stabs_bc_stab_callback,
-				  sizeof(stabs_bc_stab), 0);
-    dbgbc_stab = (stabs_bc_stab *)dbgbc;
-
+    dbgbc = yasm_bc_create_common(&stabs_bc_stab_callback, stab, 0);
     dbgbc->len = info.stablen;
-    dbgbc_stab->stab = stab;
-
     yasm_section_bcs_append(info.stab, dbgbc);
 
     /* initial strtab bytecodes */
     nullbc = stabs_dbgfmt_append_bcstr(info.stabstr, "");
-    filebc = stabs_dbgfmt_append_bcstr(info.stabstr, dbgfmt_stabs->filename);
+    filebc = stabs_dbgfmt_append_bcstr(info.stabstr,
+	yasm_object_get_source_fn(dbgfmt_stabs->object));
 
     stext = yasm_object_find_general(dbgfmt_stabs->object, ".text");
     firstsym = yasm_symtab_use(dbgfmt_stabs->symtab, ".text", 0);
@@ -425,16 +398,15 @@ stabs_dbgfmt_generate(yasm_dbgfmt *dbgfmt)
 
 static int
 stabs_bc_stab_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
-		      yasm_output_expr_func output_expr,
+		      yasm_output_value_func output_value,
 		      yasm_output_reloc_func output_reloc)
 {
     /* This entire function, essentially the core of rendering stabs to a file,
      * needs to become endian aware.  Size appears not to be an issue, as known
      * 64-bit systems use truncated values in 32-bit fields. */
 
-    stabs_bc_stab *bc_stab = (stabs_bc_stab *)bc;
+    const stabs_stab *stab = (const stabs_stab *)bc->contents;
     unsigned char *buf = *bufp;
-    const stabs_stab *stab = bc_stab->stab;
 
     YASM_WRITE_32_L(buf, stab->bcstr ? stab->bcstr->offset : 0);
     YASM_WRITE_8(buf, stab->type);
@@ -460,12 +432,11 @@ stabs_bc_stab_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
 
 static int
 stabs_bc_str_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
-		     yasm_output_expr_func output_expr,
+		     yasm_output_value_func output_value,
 		     yasm_output_reloc_func output_reloc)
 {
-    stabs_bc_str *bc_str = (stabs_bc_str *)bc;
+    const char *str = (const char *)bc->contents;
     unsigned char *buf = *bufp;
-    const char *str = bc_str->str;
 
     strcpy((char *)buf, str);
     buf += strlen(str)+1;
@@ -475,24 +446,21 @@ stabs_bc_str_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
 }
 
 static void
-stabs_bc_stab_destroy(yasm_bytecode *bc)
+stabs_bc_stab_destroy(void *contents)
 {
-    stabs_bc_stab *bc_stab = (stabs_bc_stab *)bc;
-    yasm_xfree(bc_stab->stab);
+    yasm_xfree(contents);
 }
 
 static void
-stabs_bc_str_destroy(yasm_bytecode *bc)
+stabs_bc_str_destroy(void *contents)
 {
-    stabs_bc_str *bc_str = (stabs_bc_str *)bc;
-    yasm_xfree(bc_str->str);
+    yasm_xfree(contents);
 }
 
 static void
-stabs_bc_stab_print(const yasm_bytecode *bc, FILE *f, int indent_level)
+stabs_bc_stab_print(const void *contents, FILE *f, int indent_level)
 {
-    const stabs_bc_stab *bc_stab = (const stabs_bc_stab *)bc;
-    const stabs_stab *stab = bc_stab->stab;
+    const stabs_stab *stab = (const stabs_stab *)contents;
     const char *str = "";
     fprintf(f, "%*s.stabs \"%s\", 0x%x, 0x%x, 0x%x, 0x%lx\n",
 	    indent_level, "", str, stab->type, stab->other, stab->desc,
@@ -500,10 +468,9 @@ stabs_bc_stab_print(const yasm_bytecode *bc, FILE *f, int indent_level)
 }
 
 static void
-stabs_bc_str_print(const yasm_bytecode *bc, FILE *f, int indent_level)
+stabs_bc_str_print(const void *contents, FILE *f, int indent_level)
 {
-    const stabs_bc_str *bc_str = (const stabs_bc_str *)bc;
-    fprintf(f, "%*s\"%s\"\n", indent_level, "", bc_str->str);
+    fprintf(f, "%*s\"%s\"\n", indent_level, "", (const char *)contents);
 }
 
 static yasm_bc_resolve_flags
@@ -526,14 +493,14 @@ stabs_bc_str_resolve(yasm_bytecode *bc, int save,
 
 static int
 stabs_dbgfmt_directive(yasm_dbgfmt *dbgfmt, const char *name,
-		       yasm_valparamhead *valparams, unsigned long line)
+		       yasm_section *sect, yasm_valparamhead *valparams,
+		       unsigned long line)
 {
     return 1;
 }
 
 /* Define dbgfmt structure -- see dbgfmt.h for details */
 yasm_dbgfmt_module yasm_stabs_LTX_dbgfmt = {
-    YASM_DBGFMT_VERSION,
     "Stabs debugging format",
     "stabs",
     stabs_dbgfmt_create,
