@@ -25,7 +25,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: bin-objfmt.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: bin-objfmt.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_BC_INTERNAL
@@ -35,14 +35,6 @@
 
 #define REGULAR_OUTBUF_SIZE	1024
 
-static void bin_section_data_destroy(/*@only@*/ void *d);
-static void bin_section_data_print(void *data, FILE *f, int indent_level);
-
-static const yasm_assoc_data_callback bin_section_data_callback = {
-    bin_section_data_destroy,
-    bin_section_data_print
-};
-
 typedef struct yasm_objfmt_bin {
     yasm_objfmt_base objfmt;		/* base structure */
 
@@ -55,8 +47,7 @@ yasm_objfmt_module yasm_bin_LTX_objfmt;
 
 
 static yasm_objfmt *
-bin_objfmt_create(/*@unused@*/ const char *in_filename, yasm_object *object,
-		  yasm_arch *a)
+bin_objfmt_create(yasm_object *object, yasm_arch *a)
 {
     yasm_objfmt_bin *objfmt_bin = yasm_xmalloc(sizeof(yasm_objfmt_bin));
     objfmt_bin->objfmt.module = &yasm_bin_LTX_objfmt;
@@ -66,20 +57,18 @@ bin_objfmt_create(/*@unused@*/ const char *in_filename, yasm_object *object,
     return (yasm_objfmt *)objfmt_bin;
 }
 
-/* Aligns sect to either its specified alignment (in its objfmt-specific data)
- * or def_align if no alignment was specified.  Uses prevsect and base to both
- * determine the new starting address (returned) and the total length of
+/* Aligns sect to either its specified alignment.  Uses prevsect and base to
+ * both determine the new starting address (returned) and the total length of
  * prevsect after sect has been aligned.
  */
 static unsigned long
 bin_objfmt_align_section(yasm_section *sect, yasm_section *prevsect,
-			 unsigned long base, unsigned long def_align,
+			 unsigned long base,
 			 /*@out@*/ unsigned long *prevsectlen,
 			 /*@out@*/ unsigned long *padamt)
 {
     /*@dependent@*/ /*@null@*/ yasm_bytecode *last;
     unsigned long start;
-    /*@dependent@*/ /*@null@*/ unsigned long *alignptr;
     unsigned long align;
 
     /* Figure out the size of .text by looking at the last bytecode's offset
@@ -96,11 +85,7 @@ bin_objfmt_align_section(yasm_section *sect, yasm_section *prevsect,
      * indicate padded size.  Because aignment is always a power of two, we
      * can use some bit trickery to do this easily.
      */
-    alignptr = yasm_section_get_data(sect, &bin_section_data_callback);
-    if (alignptr)
-	align = *alignptr;
-    else
-	align = def_align;	/* No alignment: use default */
+    align = yasm_section_get_align(sect);
 
     if (start & (align-1))
 	start = (start & ~(align-1)) + align;
@@ -131,7 +116,8 @@ bin_objfmt_expr_xform(/*@returned@*/ /*@only@*/ yasm_expr *e,
 	/* Transform symrecs that reference sections into
 	 * start expr + intnum(dist).
 	 */
-	if (e->terms[i].type == YASM_EXPR_SYM &&
+	if ((e->terms[i].type == YASM_EXPR_SYM ||
+	     e->terms[i].type == YASM_EXPR_SYMEXP) &&
 	    yasm_symrec_get_label(e->terms[i].data.sym, &precbc) &&
 	    (sect = yasm_bc_get_section(precbc)) &&
 	    (dist = yasm_common_calc_bc_dist(yasm_section_bcs_first(sect),
@@ -149,54 +135,46 @@ bin_objfmt_expr_xform(/*@returned@*/ /*@only@*/ yasm_expr *e,
 }
 
 static int
-bin_objfmt_output_expr(yasm_expr **ep, unsigned char *buf, size_t destsize,
-		       size_t valsize, int shift,
-		       /*@unused@*/ unsigned long offset, yasm_bytecode *bc,
-		       int rel, int warn, /*@null@*/ void *d)
+bin_objfmt_output_value(yasm_value *value, unsigned char *buf, size_t destsize,
+			size_t valsize, int shift,
+			/*@unused@*/ unsigned long offset, yasm_bytecode *bc,
+			int warn, /*@null@*/ void *d)
 {
     /*@null@*/ bin_objfmt_output_info *info = (bin_objfmt_output_info *)d;
-    /*@dependent@*/ /*@null@*/ yasm_intnum *intn;
-    /*@dependent@*/ /*@null@*/ const yasm_floatnum *flt;
+    /*@dependent@*/ /*@null@*/ yasm_bytecode *precbc;
+    /*@dependent@*/ yasm_section *sect;
 
     assert(info != NULL);
 
-    /* For binary output, this is trivial: any expression that doesn't simplify
-     * to an integer is an error (references something external).
-     * Other object formats need to generate their relocation list from here!
-     */
-
-    *ep = yasm_expr__level_tree(*ep, 1, 1, NULL, bin_objfmt_expr_xform, NULL,
-				NULL);
-
-    /* Handle floating point expressions */
-    flt = yasm_expr_get_floatnum(ep);
-    if (flt) {
-	if (shift < 0)
-	    yasm_internal_error(N_("attempting to negative shift a float"));
-	return yasm_arch_floatnum_tobytes(info->objfmt_bin->arch, flt, buf,
-					  destsize, valsize,
-					  (unsigned int)shift, warn, bc->line);
+    /* Binary objects we need to resolve against object, not against section. */
+    if (value->rel && !value->curpos_rel
+	&& yasm_symrec_get_label(value->rel, &precbc)
+	&& (sect = yasm_bc_get_section(precbc))) {
+	if (!value->abs)
+	    value->abs = yasm_expr_create_ident(yasm_expr_sym(value->rel),
+						bc->line);
+	else
+	    value->abs =
+		yasm_expr_create(YASM_EXPR_ADD, yasm_expr_expr(value->abs),
+				 yasm_expr_sym(value->rel), bc->line);
+	value->rel = NULL;
     }
 
-    /* Handle integer expressions */
-    intn = yasm_expr_get_intnum(ep, NULL);
-    if (intn) {
-	if (rel) {
-	    int retval = yasm_arch_intnum_fixup_rel(info->objfmt_bin->arch,
-						    intn, valsize, bc,
-						    bc->line);
-	    if (retval)
-		return retval;
-	}
-	return yasm_arch_intnum_tobytes(info->objfmt_bin->arch, intn, buf,
-					destsize, valsize, shift, bc, warn,
-					bc->line);
-    }
+    /* Simplify absolute portion of value, transforming symrecs */
+    if (value->abs)
+	value->abs = yasm_expr__level_tree
+	    (value->abs, 1, 1, 1, NULL, bin_objfmt_expr_xform, NULL, NULL,
+	     NULL);
 
-    /* Check for complex float expressions */
-    if (yasm_expr__contains(*ep, YASM_EXPR_FLOAT)) {
-	yasm__error(bc->line, N_("floating point expression too complex"));
-	return 1;
+    /* Output */
+    switch (yasm_value_output_basic(value, buf, destsize, valsize, shift,
+				    bc, warn, info->objfmt_bin->arch, NULL)) {
+	case -1:
+	    return 1;
+	case 0:
+	    break;
+	default:
+	    return 0;
     }
 
     /* Couldn't output, assume it contains an external reference. */
@@ -211,14 +189,12 @@ bin_objfmt_output_bytecode(yasm_bytecode *bc, /*@null@*/ void *d)
     /*@null@*/ bin_objfmt_output_info *info = (bin_objfmt_output_info *)d;
     /*@null@*/ /*@only@*/ unsigned char *bigbuf;
     unsigned long size = REGULAR_OUTBUF_SIZE;
-    unsigned long multiple;
-    unsigned long i;
     int gap;
 
     assert(info != NULL);
 
-    bigbuf = yasm_bc_tobytes(bc, info->buf, &size, &multiple, &gap, info,
-			     bin_objfmt_output_expr, NULL);
+    bigbuf = yasm_bc_tobytes(bc, info->buf, &size, &gap, info,
+			     bin_objfmt_output_value, NULL);
 
     /* Don't bother doing anything else if size ended up being 0. */
     if (size == 0) {
@@ -230,20 +206,19 @@ bin_objfmt_output_bytecode(yasm_bytecode *bc, /*@null@*/ void *d)
     /* Warn that gaps are converted to 0 and write out the 0's. */
     if (gap) {
 	unsigned long left;
-	yasm__warning(YASM_WARN_GENERAL, bc->line,
+	yasm__warning(YASM_WARN_UNINIT_CONTENTS, bc->line,
 	    N_("uninitialized space declared in code/data section: zeroing"));
 	/* Write out in chunks */
 	memset(info->buf, 0, REGULAR_OUTBUF_SIZE);
-	left = multiple*size;
+	left = size;
 	while (left > REGULAR_OUTBUF_SIZE) {
 	    fwrite(info->buf, REGULAR_OUTBUF_SIZE, 1, info->f);
 	    left -= REGULAR_OUTBUF_SIZE;
 	}
 	fwrite(info->buf, left, 1, info->f);
     } else {
-	/* Output multiple copies of buf (or bigbuf if non-NULL) to file */
-	for (i=0; i<multiple; i++)
-	    fwrite(bigbuf ? bigbuf : info->buf, (size_t)size, 1, info->f);
+	/* Output buf (or bigbuf if non-NULL) to file */
+	fwrite(bigbuf ? bigbuf : info->buf, (size_t)size, 1, info->f);
     }
 
     /* If bigbuf was allocated, free it */
@@ -254,8 +229,8 @@ bin_objfmt_output_bytecode(yasm_bytecode *bc, /*@null@*/ void *d)
 }
 
 static void
-bin_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
-		  /*@unused@*/ int all_syms, /*@unused@*/ yasm_dbgfmt *df)
+bin_objfmt_output(yasm_objfmt *objfmt, FILE *f, /*@unused@*/ int all_syms,
+		  /*@unused@*/ yasm_dbgfmt *df)
 {
     yasm_objfmt_bin *objfmt_bin = (yasm_objfmt_bin *)objfmt;
     /*@observer@*/ /*@null@*/ yasm_section *text, *data, *bss, *prevsect;
@@ -302,7 +277,7 @@ bin_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
     prevsectlenptr = &textlen;
     prevsectpadptr = &textpad;
     if (data) {
-	start = bin_objfmt_align_section(data, prevsect, start, 4,
+	start = bin_objfmt_align_section(data, prevsect, start,
 					 prevsectlenptr, prevsectpadptr);
 	yasm_section_set_start(data, yasm_expr_create_ident(
 	    yasm_expr_int(yasm_intnum_create_uint(start)), 0), 0);
@@ -312,7 +287,7 @@ bin_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
 	prevsectpadptr = &datapad;
     }
     if (bss) {
-	start = bin_objfmt_align_section(bss, prevsect, start, 4,
+	start = bin_objfmt_align_section(bss, prevsect, start,
 					 prevsectlenptr, prevsectpadptr);
 	yasm_section_set_start(bss, yasm_expr_create_ident(
 	    yasm_expr_int(yasm_intnum_create_uint(start)), 0), 0);
@@ -349,6 +324,31 @@ bin_objfmt_destroy(yasm_objfmt *objfmt)
     yasm_xfree(objfmt);
 }
 
+static void
+bin_objfmt_init_new_section(yasm_objfmt_bin *objfmt_bin, yasm_section *sect,
+			    const char *sectname, unsigned long line)
+{
+    yasm_symtab_define_label(
+	yasm_object_get_symtab(objfmt_bin->object), sectname,
+	yasm_section_bcs_first(sect), 1, line);
+}
+
+static yasm_section *
+bin_objfmt_add_default_section(yasm_objfmt *objfmt)
+{
+    yasm_objfmt_bin *objfmt_bin = (yasm_objfmt_bin *)objfmt;
+    yasm_section *retval;
+    int isnew;
+
+    retval = yasm_object_get_general(objfmt_bin->object, ".text", 0, 16, 1, 0,
+				     &isnew, 0);
+    if (isnew) {
+	bin_objfmt_init_new_section(objfmt_bin, retval, ".text", 0);
+	yasm_section_set_default(retval, 1);
+    }
+    return retval;
+}
+
 static /*@observer@*/ /*@null@*/ yasm_section *
 bin_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 			  /*@unused@*/ /*@null@*/
@@ -362,8 +362,8 @@ bin_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
     unsigned long start;
     char *sectname;
     int resonly = 0;
-    unsigned long alignval = 0;
-    int have_alignval = 0;
+    unsigned long align = 4;
+    int have_align = 0;
 
     if ((vp = yasm_vps_first(valparams)) && !vp->param && vp->val != NULL) {
 	/* If it's the first section output (.text) start at 0, otherwise
@@ -386,9 +386,14 @@ bin_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 
 	/* Check for ALIGN qualifier */
 	while ((vp = yasm_vps_next(vp))) {
+	    if (!vp->val) {
+		yasm__warning(YASM_WARN_GENERAL, line,
+			      N_("Unrecognized numeric qualifier"));
+		continue;
+	    }
+
 	    if (yasm__strcasecmp(vp->val, "align") == 0 && vp->param) {
-		/*@dependent@*/ /*@null@*/ const yasm_intnum *align;
-		unsigned long bitcnt;
+		/*@dependent@*/ /*@null@*/ const yasm_intnum *align_expr;
 
 		if (strcmp(sectname, ".text") == 0) {
 		    yasm__error(line,
@@ -397,47 +402,39 @@ bin_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 		    return NULL;
 		}
 		
-		align = yasm_expr_get_intnum(&vp->param, NULL);
-		if (!align) {
+		align_expr = yasm_expr_get_intnum(&vp->param, NULL);
+		if (!align_expr) {
 		    yasm__error(line,
 				N_("argument to `%s' is not a power of two"),
 				vp->val);
 		    return NULL;
 		}
-		alignval = yasm_intnum_get_uint(align);
+		align = yasm_intnum_get_uint(align_expr);
 
-		/* Check to see if alignval is a power of two.
-		 * This can be checked by seeing if only one bit is set.
-		 */
-		BitCount(bitcnt, alignval);
-		if (bitcnt > 1) {
+		/* Alignments must be a power of two. */
+		if ((align & (align - 1)) != 0) {
 		    yasm__error(line,
 				N_("argument to `%s' is not a power of two"),
 				vp->val);
 		    return NULL;
 		}
 
-		have_alignval = 1;
+		have_align = 1;
 	    }
 	}
 
 	retval = yasm_object_get_general(objfmt_bin->object, sectname,
 	    yasm_expr_create_ident(
-		yasm_expr_int(yasm_intnum_create_uint(start)), line), resonly,
-	    &isnew, line);
+		yasm_expr_int(yasm_intnum_create_uint(start)), line), align,
+	    strcmp(sectname, ".text") == 0, resonly, &isnew, line);
 
-	if (isnew) {
-	    if (have_alignval) {
-		unsigned long *data = yasm_xmalloc(sizeof(unsigned long));
-		*data = alignval;
-		yasm_section_add_data(retval, &bin_section_data_callback,
-				      data);
-	    }
+	if (isnew)
+	    bin_objfmt_init_new_section(objfmt_bin, retval, sectname, line);
 
-	    yasm_symtab_define_label(
-		yasm_object_get_symtab(objfmt_bin->object), sectname,
-		yasm_section_bcs_first(retval), 1, line);
-	} else if (have_alignval)
+	if (isnew || yasm_section_is_default(retval)) {
+	    yasm_section_set_default(retval, 0);
+	    yasm_section_set_align(retval, align, line);
+	} else if (have_align)
 	    yasm__warning(YASM_WARN_GENERAL, line,
 		N_("alignment value ignored on section redeclaration"));
 
@@ -446,12 +443,6 @@ bin_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 	return NULL;
 }
 
-static void
-bin_section_data_destroy(/*@only@*/ void *d)
-{
-    yasm_xfree(d);
-}
-
 static yasm_symrec *
 bin_objfmt_extern_declare(yasm_objfmt *objfmt, const char *name,
 			  /*@unused@*/ /*@null@*/
@@ -542,12 +533,6 @@ bin_objfmt_directive(yasm_objfmt *objfmt, const char *name,
 	return 1;	    /* directive unrecognized */
 }
 
-static void
-bin_section_data_print(void *data, FILE *f, int indent_level)
-{
-    fprintf(f, "%*salign=%ld\n", indent_level, "", *((unsigned long *)data));
-}
-
 
 /* Define valid debug formats to use with this object format */
 static const char *bin_objfmt_dbgfmt_keywords[] = {
@@ -557,17 +542,16 @@ static const char *bin_objfmt_dbgfmt_keywords[] = {
 
 /* Define objfmt structure -- see objfmt.h for details */
 yasm_objfmt_module yasm_bin_LTX_objfmt = {
-    YASM_OBJFMT_VERSION,
     "Flat format binary",
     "bin",
     NULL,
-    ".text",
     16,
     bin_objfmt_dbgfmt_keywords,
     "null",
     bin_objfmt_create,
     bin_objfmt_output,
     bin_objfmt_destroy,
+    bin_objfmt_add_default_section,
     bin_objfmt_section_switch,
     bin_objfmt_extern_declare,
     bin_objfmt_global_declare,
