@@ -26,7 +26,7 @@
  */
 
 #include <util.h>
-/*@unused@*/ RCSID("$Id: elf-x86-amd64.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: elf-x86-amd64.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_EXPR_INTERNAL
@@ -63,7 +63,7 @@ elf_x86_amd64_write_symtab_entry(unsigned char *bufp,
 {
     YASM_WRITE_32_L(bufp, entry->name ? entry->name->index : 0);
     YASM_WRITE_8(bufp, ELF64_ST_INFO(entry->bind, entry->type));
-    YASM_WRITE_8(bufp, 0);
+    YASM_WRITE_8(bufp, ELF64_ST_OTHER(entry->vis));
     if (entry->sect) {
         if (yasm_section_is_absolute(entry->sect)) {
             YASM_WRITE_16_L(bufp, SHN_ABS);
@@ -95,12 +95,8 @@ elf_x86_amd64_write_secthead(unsigned char *bufp, elf_secthead *shead)
     YASM_WRITE_32_L(bufp, shead->link);
     YASM_WRITE_32_L(bufp, shead->info);
 
-    if (shead->align)
-        YASM_WRITE_64I_L(bufp, shead->align);
-    else
-        YASM_WRITE_64Z_L(bufp, 0);
+    YASM_WRITE_64Z_L(bufp, shead->align);
     YASM_WRITE_64Z_L(bufp, shead->entsize);
-    
 }
 
 static void
@@ -178,7 +174,12 @@ elf_x86_amd64_write_reloc(unsigned char *bufp, elf_reloc_entry *reloc,
     YASM_WRITE_64I_L(bufp, reloc->reloc.addr);
     /*YASM_WRITE_64_L(bufp, ELF64_R_INFO(r_sym, r_type));*/
     YASM_WRITE_64C_L(bufp, r_sym, r_type);
-    YASM_WRITE_64I_L(bufp, reloc->addend);
+    if (reloc->addend)
+	YASM_WRITE_64I_L(bufp, reloc->addend);
+    else {
+	YASM_WRITE_32_L(bufp, 0);
+	YASM_WRITE_32_L(bufp, 0);
+    }
 }
 
 static void
@@ -233,5 +234,6 @@ elf_machine_handler_x86_amd64 = {
     elf_x86_amd64_write_reloc,
     elf_x86_amd64_write_proghead,
     elf_x86_amd64_ssyms,
-    sizeof(elf_x86_amd64_ssyms)/sizeof(elf_x86_amd64_ssyms[0])
+    sizeof(elf_x86_amd64_ssyms)/sizeof(elf_x86_amd64_ssyms[0]),
+    64
 };
