@@ -26,7 +26,7 @@
  */
 
 #include <util.h>
-/*@unused@*/ RCSID("$Id: elf-x86-x86.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: elf-x86-x86.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_EXPR_INTERNAL
@@ -68,7 +68,7 @@ elf_x86_x86_write_symtab_entry(unsigned char *bufp,
     YASM_WRITE_32I_L(bufp, size_intn);
 
     YASM_WRITE_8(bufp, ELF32_ST_INFO(entry->bind, entry->type));
-    YASM_WRITE_8(bufp, 0);
+    YASM_WRITE_8(bufp, ELF32_ST_OTHER(entry->vis));
     if (entry->sect) {
         if (yasm_section_is_absolute(entry->sect)) {
             YASM_WRITE_16_L(bufp, SHN_ABS);
@@ -98,10 +98,7 @@ elf_x86_x86_write_secthead(unsigned char *bufp, elf_secthead *shead)
     YASM_WRITE_32_L(bufp, shead->link);
     YASM_WRITE_32_L(bufp, shead->info);
 
-    if (shead->align)
-        YASM_WRITE_32I_L(bufp, shead->align);
-    else
-        YASM_WRITE_32_L(bufp, 0);
+    YASM_WRITE_32_L(bufp, shead->align);
     YASM_WRITE_32_L(bufp, shead->entsize);
 
 }
@@ -170,7 +167,7 @@ elf_x86_x86_write_reloc(unsigned char *bufp, elf_reloc_entry *reloc,
                         unsigned int r_type, unsigned int r_sym)
 {
     YASM_WRITE_32I_L(bufp, reloc->reloc.addr);
-    YASM_WRITE_32_L(bufp, ELF32_R_INFO((unsigned char)r_sym, (unsigned char)r_type));
+    YASM_WRITE_32_L(bufp, ELF32_R_INFO((unsigned long)r_sym, (unsigned char)r_type));
 }
 
 static void
@@ -223,5 +220,6 @@ elf_machine_handler_x86_x86 = {
     elf_x86_x86_write_reloc,
     elf_x86_x86_write_proghead,
     elf_x86_x86_ssyms,
-    sizeof(elf_x86_x86_ssyms)/sizeof(elf_x86_x86_ssyms[0])
+    sizeof(elf_x86_x86_ssyms)/sizeof(elf_x86_x86_ssyms[0]),
+    32
 };
