@@ -26,7 +26,7 @@
  */
 
 #include <util.h>
-/*@unused@*/ RCSID("$Id: elf.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: elf.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_EXPR_INTERNAL
@@ -61,12 +61,14 @@ static const elf_machine_handler *elf_machine_handlers[] =
     &elf_machine_handler_x86_amd64,
     NULL
 };
-static const elf_machine_handler elf_null_machine = {0};
+static const elf_machine_handler elf_null_machine = {0, 0, 0, 0, 0, 0, 0, 0,
+						     0, 0, 0, 0, 0, 0, 0, 0,
+						     0, 0, 0};
 static elf_machine_handler const *elf_march = &elf_null_machine;
 static yasm_symrec **elf_ssyms;
 
-int
-elf_set_arch(yasm_arch *arch, yasm_symtab *symtab)
+const elf_machine_handler *
+elf_set_arch(yasm_arch *arch, yasm_symtab *symtab, int bits_pref)
 {
     const char *machine = yasm_arch_get_machine(arch);
     int i;
@@ -77,7 +79,8 @@ elf_set_arch(yasm_arch *arch, yasm_symtab *symtab)
     {
         if (yasm__strcasecmp(yasm_arch_keyword(arch), elf_march->arch)==0)
             if (yasm__strcasecmp(machine, elf_march->machine)==0)
-                break;
+		if (bits_pref == 0 || bits_pref == elf_march->bits)
+		    break;
     }
 
     if (elf_march && elf_march->num_ssyms > 0)
@@ -94,7 +97,7 @@ elf_set_arch(yasm_arch *arch, yasm_symtab *symtab)
 	}
     }
 
-    return elf_march != NULL;
+    return elf_march;
 }
 
 /* reloc functions */
@@ -161,6 +164,24 @@ elf_strtab_entry_create(const char *str)
     return entry;
 }
 
+void
+elf_strtab_entry_set_str(elf_strtab_entry *entry, const char *str)
+{
+    elf_strtab_entry *last;
+    if (entry->str)
+	yasm_xfree(entry->str);
+    entry->str = yasm__xstrdup(str);
+
+    /* Update all following indices since string length probably changes */
+    last = entry;
+    entry = STAILQ_NEXT(last, qlink);
+    while (entry) {
+	entry->index = last->index + strlen(last->str) + 1;
+	last = entry;
+	entry = STAILQ_NEXT(last, qlink);
+    }
+}
+
 elf_strtab_head *
 elf_strtab_create()
 {
@@ -240,6 +261,7 @@ elf_symtab_entry_create(elf_strtab_entry *name,
 			yasm_symrec *sym)
 {
     elf_symtab_entry *entry = yasm_xmalloc(sizeof(elf_symtab_entry));
+    entry->in_table = 0;
     entry->sym = sym;
     entry->sect = NULL;
     entry->name = name;
@@ -250,6 +272,7 @@ elf_symtab_entry_create(elf_strtab_entry *name,
     entry->index = 0;
     entry->bind = 0;
     entry->type = STT_NOTYPE;
+    entry->vis = STV_DEFAULT;
 
     return entry;
 }
@@ -313,6 +336,7 @@ elf_symtab_create()
     elf_symtab_entry *entry = yasm_xmalloc(sizeof(elf_symtab_entry));
 
     STAILQ_INIT(symtab);
+    entry->in_table = 1;
     entry->sym = NULL;
     entry->sect = NULL;
     entry->name = NULL;
@@ -320,14 +344,15 @@ elf_symtab_create()
     entry->xsize = NULL;
     entry->size = 0;
     entry->index = SHN_UNDEF;
-    entry->bind = 0;
-    entry->type = 0;
+    entry->bind = STB_LOCAL;
+    entry->type = STT_NOTYPE;
+    entry->vis = STV_DEFAULT;
     entry->symindex = 0;
     STAILQ_INSERT_TAIL(symtab, entry, qlink);
     return symtab;
 }
 
-elf_symtab_entry *
+void
 elf_symtab_append_entry(elf_symtab_head *symtab, elf_symtab_entry *entry)
 {
     if (symtab == NULL)
@@ -338,18 +363,12 @@ elf_symtab_append_entry(elf_symtab_head *symtab, elf_symtab_entry *entry)
 	yasm_internal_error(N_("symtab is missing initial dummy entry"));
 
     STAILQ_INSERT_TAIL(symtab, entry, qlink);
-    return entry;
+    entry->in_table = 1;
 }
 
-elf_symtab_entry *
-elf_symtab_insert_local_sym(elf_symtab_head *symtab,
-			    elf_strtab_head *strtab,
-			    yasm_symrec *sym)
+void
+elf_symtab_insert_local_sym(elf_symtab_head *symtab, elf_symtab_entry *entry)
 {
-    elf_strtab_entry *name = strtab
-	? elf_strtab_append_str(strtab, yasm_symrec_get_name(sym))
-	: NULL;
-    elf_symtab_entry *entry = elf_symtab_entry_create(name, sym);
     elf_symtab_entry *after = STAILQ_FIRST(symtab);
     elf_symtab_entry *before = NULL;
 
@@ -359,8 +378,7 @@ elf_symtab_insert_local_sym(elf_symtab_head *symtab,
 	after = STAILQ_NEXT(after, qlink);
     }
     STAILQ_INSERT_AFTER(symtab, before, entry, qlink);
-
-    return entry;
+    entry->in_table = 1;
 }
 
 void
@@ -475,7 +493,7 @@ void elf_symtab_set_nonzero(elf_symtab_entry *entry,
 			    elf_symbol_binding bind,
 			    elf_symbol_type type,
 			    yasm_expr *xsize,
-			    elf_address value)
+			    elf_address *value)
 {
     if (!entry)
 	yasm_internal_error("NULL entry");
@@ -484,15 +502,42 @@ void elf_symtab_set_nonzero(elf_symtab_entry *entry,
     if (bind) entry->bind = bind;
     if (type) entry->type = type;
     if (xsize) entry->xsize = xsize;
-    if (value) entry->value = value;
+    if (value) entry->value = *value;
 }
 
+void
+elf_sym_set_visibility(elf_symtab_entry *entry,
+                       elf_symbol_vis    vis)
+{
+    entry->vis = ELF_ST_VISIBILITY(vis);
+}                            
+
+void
+elf_sym_set_type(elf_symtab_entry *entry,
+                 elf_symbol_type   type)
+{
+    entry->type = type;
+}                            
+
+void
+elf_sym_set_size(elf_symtab_entry *entry,
+                 struct yasm_expr *size)
+{
+    if (entry->xsize)
+	yasm_expr_destroy(entry->xsize);
+    entry->xsize = size;
+}                            
+
+int
+elf_sym_in_table(elf_symtab_entry *entry)
+{
+    return entry->in_table;
+}
 
 elf_secthead *
 elf_secthead_create(elf_strtab_entry	*name,
 		    elf_section_type	 type,
 		    elf_section_flags	 flags,
-		    elf_section_index	 idx,
 		    elf_address		 offset,
 		    elf_size		 size)
 {
@@ -504,15 +549,15 @@ elf_secthead_create(elf_strtab_entry	*name,
     esd->size = yasm_intnum_create_uint(size);
     esd->link = 0;
     esd->info = 0;
-    esd->align = NULL;
+    esd->align = 0;
     esd->entsize = 0;
-    esd->index = idx;
+    esd->index = 0;
 
     esd->sym = NULL;
     esd->name = name;
     esd->index = 0;
     esd->rel_name = NULL;
-    esd->rel_index = idx;
+    esd->rel_index = 0;
     esd->rel_offset = 0;
     esd->nreloc = 0;
 
@@ -520,7 +565,7 @@ elf_secthead_create(elf_strtab_entry	*name,
         if (!elf_march->symtab_entry_size || !elf_march->symtab_entry_align)
 	    yasm_internal_error(N_("unsupported ELF format"));
         esd->entsize = elf_march->symtab_entry_size;
-        esd->align = yasm_intnum_create_uint(elf_march->symtab_entry_align);
+        esd->align = elf_march->symtab_entry_align;
     }
 
     return esd;
@@ -532,8 +577,7 @@ elf_secthead_destroy(elf_secthead *shead)
     if (shead == NULL)
 	yasm_internal_error(N_("shead is null"));
 
-    if (shead->align)
-	yasm_intnum_destroy(shead->align);
+    yasm_intnum_destroy(shead->size);
 
     yasm_xfree(shead);
 }
@@ -566,8 +610,7 @@ elf_secthead_print(void *data, FILE *f, int indent_level)
     fprintf(f, "%*ssize=0x%lx\n", indent_level, "",
 	    yasm_intnum_get_uint(sect->size));
     fprintf(f, "%*slink=0x%x\n", indent_level, "", sect->link);
-    fprintf(f, "%*salign=%ld\n", indent_level, "",
-	    yasm_intnum_get_uint(sect->align));
+    fprintf(f, "%*salign=%lu\n", indent_level, "", sect->align);
     fprintf(f, "%*snreloc=%ld\n", indent_level, "", sect->nreloc);
 }
 
@@ -590,12 +633,10 @@ elf_secthead_write_to_file(FILE *f, elf_secthead *shead,
     return 0;
 }
 
-int
+void
 elf_secthead_append_reloc(yasm_section *sect, elf_secthead *shead,
 			  elf_reloc_entry *reloc)
 {
-    int new_sect = 0;
-
     if (sect == NULL)
 	yasm_internal_error("sect is null");
     if (shead == NULL)
@@ -605,8 +646,6 @@ elf_secthead_append_reloc(yasm_section *sect, elf_secthead *shead,
 
     shead->nreloc++;
     yasm_section_add_reloc(sect, (yasm_reloc *)reloc, elf_reloc_entry_destroy);
-
-    return new_sect;
 }
 
 char *
@@ -720,6 +759,14 @@ elf_secthead_get_type(elf_secthead *shead)
     return shead->type;
 }
 
+void
+elf_secthead_set_typeflags(elf_secthead *shead, elf_section_type type,
+			   elf_section_flags flags)
+{
+    shead->type = type;
+    shead->flags = flags;
+}
+
 int
 elf_secthead_is_empty(elf_secthead *shead)
 {
@@ -738,12 +785,15 @@ elf_secthead_get_index(elf_secthead *shead)
     return shead->index;
 }
 
-const yasm_intnum *
-elf_secthead_set_align(elf_secthead *shead, yasm_intnum *align)
+unsigned long
+elf_secthead_get_align(const elf_secthead *shead)
+{
+    return shead->align;
+}
+
+unsigned long
+elf_secthead_set_align(elf_secthead *shead, unsigned long align)
 {
-    if (shead->align != NULL)
-	yasm_intnum_destroy(shead->align);
-    
     return shead->align = align;
 }
 
@@ -800,7 +850,7 @@ elf_secthead_add_size(elf_secthead *shead, yasm_intnum *size)
 long
 elf_secthead_set_file_offset(elf_secthead *shead, long pos)
 {
-    unsigned long align = yasm_intnum_get_uint(shead->align);
+    unsigned long align = shead->align;
 
     if (align == 0 || align == 1) {
 	shead->offset = (unsigned long)pos;
