@@ -1,4 +1,4 @@
-/* $Id: elf.h 1168 2004-10-31 01:07:52Z peter $
+/* $Id: elf.h 1591 2006-07-13 03:25:38Z peter $
  * ELF object format helpers
  *
  *  Copyright (C) 2003  Michael Urman
@@ -36,6 +36,8 @@ typedef struct elf_strtab_head elf_strtab_head;
 typedef struct elf_symtab_entry elf_symtab_entry;
 typedef struct elf_symtab_head elf_symtab_head;
 
+typedef struct elf_machine_handler elf_machine_handler;
+
 typedef unsigned long	elf_address;
 typedef unsigned long	elf_offset;
 typedef unsigned long	elf_size;
@@ -149,6 +151,10 @@ typedef enum {
     SHF_WRITE = 0x1,		/* data should be writable at runtime */
     SHF_ALLOC = 0x2,		/* occupies memory at runtime */
     SHF_EXECINSTR = 0x4,	/* contains machine instructions */
+    SHF_MERGE = 0x10,		/* data can be merged */
+    SHF_STRINGS = 0x20,		/* contains 0-terminated strings */
+    SHF_GROUP = 0x200,		/* member of a section group */
+    SHF_TLS = 0x400,		/* thread local storage */
     SHF_MASKOS = 0x0f000000/*,*//* environment specific use */
     /*SHF_MASKPROC = 0xf0000000*/	/* bits reserved for processor specific needs */
 } elf_section_flags;
@@ -196,15 +202,28 @@ typedef enum {
     STN_UNDEF = 0
 } elf_symbol_index;
 
+/* elf symbol visibility - lower two bits of OTHER field */
+typedef enum {
+    STV_DEFAULT = 0,		/* Default symbol visibility rules */
+    STV_INTERNAL = 1,		/* Processor specific hidden class */
+    STV_HIDDEN = 2,		/* Sym unavailable in other modules */
+    STV_PROTECTED = 3		/* Not preemptable, not exported */
+} elf_symbol_vis;
+
 
 /* internal only object definitions */
 #ifdef YASM_OBJFMT_ELF_INTERNAL
 
+#define ELF_VISIBILITY_MASK		0x03
+#define ELF_ST_VISIBILITY(v)            ((v) & ELF_VISIBILITY_MASK)
+
 #define ELF32_ST_INFO(bind, type)	(((bind) << 4) + ((type) & 0xf))
 #define ELF32_R_INFO(s,t)		(((s)<<8)+(unsigned char)(t))
+#define ELF32_ST_OTHER(vis)             ELF_ST_VISIBILITY(vis)
 
 #define ELF64_ST_INFO(bind, type)	(((bind) << 4) + ((type) & 0xf))
 #define ELF64_R_INFO(s,t)		(((s)<<32) + ((t) & 0xffffffffL))
+#define ELF64_ST_OTHER(vis)             ELF_ST_VISIBILITY(vis)
 
 #define EHDR32_SIZE 52
 #define EHDR64_SIZE 64
@@ -318,7 +337,7 @@ struct elf_secthead {
     yasm_intnum		*size;
     elf_section_index	 link;
     elf_section_info	 info;	    /* see note ESD1 */
-    yasm_intnum 	*align;
+    unsigned long	 align;
     elf_size		 entsize;
 
     yasm_symrec		*sym;
@@ -367,6 +386,7 @@ struct elf_strtab_entry {
 STAILQ_HEAD(elf_symtab_head, elf_symtab_entry);
 struct elf_symtab_entry {
     STAILQ_ENTRY(elf_symtab_entry) qlink;
+    int			in_table;
     yasm_symrec		*sym;
     yasm_section	*sect;
     elf_strtab_entry	*name;
@@ -376,6 +396,7 @@ struct elf_symtab_entry {
     elf_section_index	 index;
     elf_symbol_binding	 bind;
     elf_symbol_type	 type;
+    elf_symbol_vis       vis;
     elf_symbol_index	 symindex;
 };
 
@@ -385,7 +406,9 @@ extern const yasm_assoc_data_callback elf_section_data;
 extern const yasm_assoc_data_callback elf_symrec_data;
 
 
-int elf_set_arch(struct yasm_arch *arch, yasm_symtab *symtab);
+const elf_machine_handler *elf_set_arch(struct yasm_arch *arch,
+					yasm_symtab *symtab,
+					int bits_pref);
 
 /* reloc functions */
 int elf_is_wrt_sym_relative(yasm_symrec *wrt);
@@ -398,6 +421,7 @@ void elf_reloc_entry_destroy(void *entry);
 
 /* strtab functions */
 elf_strtab_entry *elf_strtab_entry_create(const char *str);
+void elf_strtab_entry_set_str(elf_strtab_entry *entry, const char *str);
 elf_strtab_head *elf_strtab_create(void);
 elf_strtab_entry *elf_strtab_append_str(elf_strtab_head *head, const char *str);
 void elf_strtab_destroy(elf_strtab_head *head);
@@ -407,11 +431,9 @@ unsigned long elf_strtab_output_to_file(FILE *f, elf_strtab_head *head);
 elf_symtab_entry *elf_symtab_entry_create(elf_strtab_entry *name,
 					  struct yasm_symrec *sym);
 elf_symtab_head *elf_symtab_create(void);
-elf_symtab_entry *elf_symtab_append_entry(elf_symtab_head *symtab,
-					  elf_symtab_entry *entry);
-elf_symtab_entry *elf_symtab_insert_local_sym(elf_symtab_head *symtab,
-					      elf_strtab_head *strtab,
-					      struct yasm_symrec *sym);
+void elf_symtab_append_entry(elf_symtab_head *symtab, elf_symtab_entry *entry);
+void elf_symtab_insert_local_sym(elf_symtab_head *symtab,
+				 elf_symtab_entry *entry);
 void elf_symtab_destroy(elf_symtab_head *head);
 unsigned long elf_symtab_assign_indices(elf_symtab_head *symtab);
 unsigned long elf_symtab_write_to_file(FILE *f, elf_symtab_head *symtab);
@@ -421,26 +443,31 @@ void elf_symtab_set_nonzero(elf_symtab_entry	*entry,
 			    elf_symbol_binding	 bind,
 			    elf_symbol_type	 type,
 			    struct yasm_expr	*size,
-			    elf_address		 value);
-
+			    elf_address		*value);
+void elf_sym_set_visibility(elf_symtab_entry    *entry,
+                            elf_symbol_vis       vis);
+void elf_sym_set_type(elf_symtab_entry *entry, elf_symbol_type type);
+void elf_sym_set_size(elf_symtab_entry *entry, struct yasm_expr *size);
+int elf_sym_in_table(elf_symtab_entry *entry);
 
 /* section header functions */
 elf_secthead *elf_secthead_create(elf_strtab_entry	*name,
 				  elf_section_type	type,
 				  elf_section_flags	flags,
-				  elf_section_index	idx,
 				  elf_address		offset,
 				  elf_size		size);
 void elf_secthead_destroy(elf_secthead *esd);
 unsigned long elf_secthead_write_to_file(FILE *f, elf_secthead *esd,
 					 elf_section_index sindex);
-int elf_secthead_append_reloc(yasm_section *sect, elf_secthead *shead,
-			      elf_reloc_entry *reloc);
+void elf_secthead_append_reloc(yasm_section *sect, elf_secthead *shead,
+			       elf_reloc_entry *reloc);
 elf_section_type elf_secthead_get_type(elf_secthead *shead);
+void elf_secthead_set_typeflags(elf_secthead *shead, elf_section_type type,
+				elf_section_flags flags);
 int elf_secthead_is_empty(elf_secthead *shead);
 struct yasm_symrec *elf_secthead_get_sym(elf_secthead *shead);
-const struct yasm_intnum *elf_secthead_set_align(elf_secthead *shead,
-						 struct yasm_intnum *align);
+unsigned long elf_secthead_get_align(const elf_secthead *shead);
+unsigned long elf_secthead_set_align(elf_secthead *shead, unsigned long align);
 elf_section_index elf_secthead_get_index(elf_secthead *shead);
 elf_section_info elf_secthead_set_info(elf_secthead *shead,
 				       elf_section_info info);
