@@ -25,7 +25,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: xdf-objfmt.c 1168 2004-10-31 01:07:52Z peter $");
+/*@unused@*/ RCSID("$Id: xdf-objfmt.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #define YASM_BC_INTERNAL
@@ -35,7 +35,7 @@
 
 #define REGULAR_OUTBUF_SIZE	1024
 
-#define XDF_MAGIC	0x87654321
+#define XDF_MAGIC	0x87654322
 
 #define XDF_SYM_EXTERN	1
 #define XDF_SYM_GLOBAL	2
@@ -64,8 +64,8 @@ typedef struct xdf_reloc {
 typedef struct xdf_section_data {
     /*@dependent@*/ yasm_symrec *sym;	/* symbol created for this section */
     yasm_intnum *addr;	    /* starting memory address */
+    yasm_intnum *vaddr;	    /* starting virtual address */
     long scnum;		    /* section number (0=first section) */
-    unsigned int align;	    /* section alignment (0-4096) */
     enum {
 	XDF_SECT_ABSOLUTE = 0x01,
 	XDF_SECT_FLAT = 0x02,
@@ -86,17 +86,10 @@ typedef struct xdf_symrec_data {
     unsigned long index;		/* assigned XDF symbol table index */
 } xdf_symrec_data;
 
-typedef struct xdf_symtab_entry {
-    STAILQ_ENTRY(xdf_symtab_entry) link;
-    /*@dependent@*/ yasm_symrec *sym;
-} xdf_symtab_entry;
-typedef STAILQ_HEAD(xdf_symtab_head, xdf_symtab_entry) xdf_symtab_head;
-
 typedef struct yasm_objfmt_xdf {
     yasm_objfmt_base objfmt;		    /* base structure */
 
     long parse_scnum;		    /* sect numbering in parser */
-    xdf_symtab_head xdf_symtab;	    /* symbol table of indexed syms */
 
     yasm_object *object;
     yasm_symtab *symtab;
@@ -109,6 +102,10 @@ typedef struct xdf_objfmt_output_info {
     /*@only@*/ unsigned char *buf;
     yasm_section *sect;
     /*@dependent@*/ xdf_section_data *xsd;
+
+    unsigned long indx;		    /* current symbol index */
+    int all_syms;		    /* outputting all symbols? */
+    unsigned long strtab_offset;    /* current string table offset */
 } xdf_objfmt_output_info;
 
 static void xdf_section_data_destroy(/*@only@*/ void *d);
@@ -130,46 +127,8 @@ static const yasm_assoc_data_callback xdf_symrec_data_cb = {
 yasm_objfmt_module yasm_xdf_LTX_objfmt;
 
 
-static /*@dependent@*/ xdf_symtab_entry *
-xdf_objfmt_symtab_append(yasm_objfmt_xdf *objfmt_xdf, yasm_symrec *sym)
-{
-    /*@null@*/ /*@dependent@*/ xdf_symrec_data *sym_data_prev;
-    xdf_symrec_data *sym_data;
-    xdf_symtab_entry *entry;
-    unsigned long indx;
-
-    if (STAILQ_EMPTY(&objfmt_xdf->xdf_symtab))
-	indx = 0;
-    else {
-	entry = STAILQ_LAST(&objfmt_xdf->xdf_symtab, xdf_symtab_entry, link);
-	sym_data_prev = yasm_symrec_get_data(entry->sym, &xdf_symrec_data_cb);
-	assert(sym_data_prev != NULL);
-	indx = sym_data_prev->index + 1;
-    }
-
-    sym_data = yasm_xmalloc(sizeof(xdf_symrec_data));
-    sym_data->index = indx;
-    yasm_symrec_add_data(sym, &xdf_symrec_data_cb, sym_data);
-
-    entry = yasm_xmalloc(sizeof(xdf_symtab_entry));
-    entry->sym = sym;
-    STAILQ_INSERT_TAIL(&objfmt_xdf->xdf_symtab, entry, link);
-
-    return entry;
-}
-
-static int
-xdf_objfmt_append_local_sym(yasm_symrec *sym, /*@unused@*/ /*@null@*/ void *d)
-{
-    /*@null@*/ yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)d;
-    assert(objfmt_xdf != NULL);
-    if (!yasm_symrec_get_data(sym, &xdf_symrec_data_cb))
-	xdf_objfmt_symtab_append(objfmt_xdf, sym);
-    return 1;
-}
-
 static yasm_objfmt *
-xdf_objfmt_create(const char *in_filename, yasm_object *object, yasm_arch *a)
+xdf_objfmt_create(yasm_object *object, yasm_arch *a)
 {
     yasm_objfmt_xdf *objfmt_xdf = yasm_xmalloc(sizeof(yasm_objfmt_xdf));
 
@@ -191,7 +150,6 @@ xdf_objfmt_create(const char *in_filename, yasm_object *object, yasm_arch *a)
     }
 
     objfmt_xdf->parse_scnum = 0;    /* section numbering starts at 0 */
-    STAILQ_INIT(&objfmt_xdf->xdf_symtab);
 
     objfmt_xdf->objfmt.module = &yasm_xdf_LTX_objfmt;
 
@@ -199,110 +157,90 @@ xdf_objfmt_create(const char *in_filename, yasm_object *object, yasm_arch *a)
 }
 
 static int
-xdf_objfmt_output_expr(yasm_expr **ep, unsigned char *buf, size_t destsize,
+xdf_objfmt_output_value(yasm_value *value, unsigned char *buf, size_t destsize,
 			size_t valsize, int shift, unsigned long offset,
-			yasm_bytecode *bc, int rel, int warn,
-			/*@null@*/ void *d)
+			yasm_bytecode *bc, int warn, /*@null@*/ void *d)
 {
     /*@null@*/ xdf_objfmt_output_info *info = (xdf_objfmt_output_info *)d;
     yasm_objfmt_xdf *objfmt_xdf;
     /*@dependent@*/ /*@null@*/ yasm_intnum *intn;
-    /*@dependent@*/ /*@null@*/ const yasm_floatnum *flt;
-    /*@dependent@*/ /*@null@*/ yasm_symrec *sym;
-    yasm_expr *shr_expr;
-    yasm_expr *wrt_expr;
-    unsigned int shr = 0;
-    unsigned int seg = 0;
+    unsigned long intn_minus;
+    int retval;
 
     assert(info != NULL);
     objfmt_xdf = info->objfmt_xdf;
 
-    *ep = yasm_expr_simplify(*ep, yasm_common_calc_bc_dist);
+    if (value->abs)
+	value->abs = yasm_expr_simplify(value->abs, yasm_common_calc_bc_dist);
 
-    /* Handle floating point expressions */
-    flt = yasm_expr_get_floatnum(ep);
-    if (flt) {
-	if (shift < 0)
-	    yasm_internal_error(N_("attempting to negative shift a float"));
-	return yasm_arch_floatnum_tobytes(objfmt_xdf->arch, flt, buf,
-					  destsize, valsize,
-					  (unsigned int)shift, warn, bc->line);
-    }
-
-    /* Check for a right shift value */
-    shr_expr = yasm_expr_extract_shr(ep);
-    if (shr_expr) {
-	/*@dependent@*/ /*@null@*/ const yasm_intnum *shr_intn;
-	shr_intn = yasm_expr_get_intnum(&shr_expr, NULL);
-	if (!shr_intn) {
-	    yasm__error(bc->line, N_("shift expression too complex"));
+    /* Try to output constant and PC-relative section-local first.
+     * Note this does NOT output any value with a SEG, WRT, external,
+     * cross-section, or non-PC-relative reference (those are handled below).
+     */
+    switch (yasm_value_output_basic(value, buf, destsize, valsize, shift, bc,
+				    warn, info->objfmt_xdf->arch,
+				    yasm_common_calc_bc_dist)) {
+	case -1:
 	    return 1;
-	}
-	shr = yasm_intnum_get_uint(shr_intn);
+	case 0:
+	    break;
+	default:
+	    return 0;
     }
 
-    /* Check for a segment relocation */
-    if (yasm_expr_extract_seg(ep))
-	seg = 1;
+    if (value->section_rel) {
+	yasm__error(bc->line, N_("xdf: relocation too complex"));
+	return 1;
+    }
 
-    /* Check for a WRT relocation */
-    wrt_expr = yasm_expr_extract_wrt(ep);
-
-    /* Handle integer expressions, with relocation if necessary */
-    sym = yasm_expr_extract_symrec(ep, 0, yasm_common_calc_bc_dist);
-    if (sym) {
+    intn_minus = 0;
+    if (value->rel) {
 	xdf_reloc *reloc;
 
 	reloc = yasm_xmalloc(sizeof(xdf_reloc));
 	reloc->reloc.addr = yasm_intnum_create_uint(bc->offset + offset);
-	reloc->reloc.sym = sym;
+	reloc->reloc.sym = value->rel;
 	reloc->base = NULL;
 	reloc->size = valsize/8;
-	reloc->shift = shr;
+	reloc->shift = value->rshift;
 
-	if (seg)
+	if (value->seg_of)
 	    reloc->type = XDF_RELOC_SEG;
-	else if (wrt_expr) {
-	    reloc->base = yasm_expr_extract_symrec(&wrt_expr, 0,
-						   yasm_common_calc_bc_dist);
-	    if (!reloc->base) {
-		yasm__error(bc->line, N_("WRT expression too complex"));
-		return 1;
-	    }
+	else if (value->wrt) {
+	    reloc->base = value->wrt;
 	    reloc->type = XDF_RELOC_WRT;
-	} else if (rel) {
+	} else if (value->curpos_rel) {
 	    reloc->type = XDF_RELOC_RIP;
-	    /* Need to reference to start of section, so add $$ in. */
-	    *ep = yasm_expr_create(YASM_EXPR_ADD, yasm_expr_expr(*ep),
-		yasm_expr_sym(yasm_symtab_define_label2("$$",
-		    yasm_section_bcs_first(info->sect), 0, (*ep)->line)),
-		(*ep)->line);
-	    *ep = yasm_expr_simplify(*ep, yasm_common_calc_bc_dist);
+	    /* Adjust to start of section, so subtract out the bytecode
+	     * offset.
+	     */
+	    intn_minus = bc->offset;
 	} else
 	    reloc->type = XDF_RELOC_REL;
 	info->xsd->nreloc++;
 	yasm_section_add_reloc(info->sect, (yasm_reloc *)reloc, yasm_xfree);
     }
-    intn = yasm_expr_get_intnum(ep, NULL);
-    if (intn) {
-	if (rel) {
-	    int retval = yasm_arch_intnum_fixup_rel(objfmt_xdf->arch, intn,
-						    valsize, bc, bc->line);
-	    if (retval)
-		return retval;
+
+    if (intn_minus > 0) {
+	intn = yasm_intnum_create_uint(intn_minus);
+	yasm_intnum_calc(intn, YASM_EXPR_NEG, NULL, bc->line);
+    } else
+	intn = yasm_intnum_create_uint(0);
+
+    if (value->abs) {
+	yasm_intnum *intn2 = yasm_expr_get_intnum(&value->abs, NULL);
+	if (!intn2) {
+	    yasm__error(bc->line, N_("xdf: relocation too complex"));
+	    yasm_intnum_destroy(intn);
+	    return 1;
 	}
-	return yasm_arch_intnum_tobytes(objfmt_xdf->arch, intn, buf, destsize,
-					valsize, shift, bc, warn, bc->line);
+	yasm_intnum_calc(intn, YASM_EXPR_ADD, intn2, bc->line);
     }
 
-    /* Check for complex float expressions */
-    if (yasm_expr__contains(*ep, YASM_EXPR_FLOAT)) {
-	yasm__error(bc->line, N_("floating point expression too complex"));
-	return 1;
-    }
-
-    yasm__error(bc->line, N_("xdf: relocation too complex"));
-    return 1;
+    retval = yasm_arch_intnum_tobytes(objfmt_xdf->arch, intn, buf, destsize,
+				      valsize, shift, bc, warn, bc->line);
+    yasm_intnum_destroy(intn);
+    return retval;
 }
 
 static int
@@ -311,14 +249,12 @@ xdf_objfmt_output_bytecode(yasm_bytecode *bc, /*@null@*/ void *d)
     /*@null@*/ xdf_objfmt_output_info *info = (xdf_objfmt_output_info *)d;
     /*@null@*/ /*@only@*/ unsigned char *bigbuf;
     unsigned long size = REGULAR_OUTBUF_SIZE;
-    unsigned long multiple;
-    unsigned long i;
     int gap;
 
     assert(info != NULL);
 
-    bigbuf = yasm_bc_tobytes(bc, info->buf, &size, &multiple, &gap, info,
-			     xdf_objfmt_output_expr, NULL);
+    bigbuf = yasm_bc_tobytes(bc, info->buf, &size, &gap, info,
+			     xdf_objfmt_output_value, NULL);
 
     /* Don't bother doing anything else if size ended up being 0. */
     if (size == 0) {
@@ -327,25 +263,24 @@ xdf_objfmt_output_bytecode(yasm_bytecode *bc, /*@null@*/ void *d)
 	return 0;
     }
 
-    info->xsd->size += multiple*size;
+    info->xsd->size += size;
 
     /* Warn that gaps are converted to 0 and write out the 0's. */
     if (gap) {
 	unsigned long left;
-	yasm__warning(YASM_WARN_GENERAL, bc->line,
+	yasm__warning(YASM_WARN_UNINIT_CONTENTS, bc->line,
 	    N_("uninitialized space: zeroing"));
 	/* Write out in chunks */
 	memset(info->buf, 0, REGULAR_OUTBUF_SIZE);
-	left = multiple*size;
+	left = size;
 	while (left > REGULAR_OUTBUF_SIZE) {
 	    fwrite(info->buf, REGULAR_OUTBUF_SIZE, 1, info->f);
 	    left -= REGULAR_OUTBUF_SIZE;
 	}
 	fwrite(info->buf, left, 1, info->f);
     } else {
-	/* Output multiple copies of buf (or bigbuf if non-NULL) to file */
-	for (i=0; i<multiple; i++)
-	    fwrite(bigbuf ? bigbuf : info->buf, (size_t)size, 1, info->f);
+	/* Output buf (or bigbuf if non-NULL) to file */
+	fwrite(bigbuf ? bigbuf : info->buf, (size_t)size, 1, info->f);
     }
 
     /* If bigbuf was allocated, free it */
@@ -482,84 +417,68 @@ xdf_objfmt_output_secthead(yasm_section *sect, /*@null@*/ void *d)
 	YASM_WRITE_32_L(localbuf, 0);
 	YASM_WRITE_32_L(localbuf, 0);
     }
-    YASM_WRITE_16_L(localbuf, xsd->align);	/* alignment */
+    if (xsd->vaddr) {
+	yasm_intnum_get_sized(xsd->vaddr, localbuf, 8, 64, 0, 0, 0, 0);
+	localbuf += 8;				/* virtual address */
+    } else if (xsd->addr) {
+	yasm_intnum_get_sized(xsd->addr, localbuf, 8, 64, 0, 0, 0, 0);
+	localbuf += 8;				/* VA=PA */
+    } else {
+	YASM_WRITE_32_L(localbuf, 0);
+	YASM_WRITE_32_L(localbuf, 0);
+    }
+    YASM_WRITE_16_L(localbuf, yasm_section_get_align(sect)); /* alignment */
     YASM_WRITE_16_L(localbuf, xsd->flags);	/* flags */
     YASM_WRITE_32_L(localbuf, xsd->scnptr);	/* file ptr to data */
     YASM_WRITE_32_L(localbuf, xsd->size);	/* section size */
     YASM_WRITE_32_L(localbuf, xsd->relptr);	/* file ptr to relocs */
     YASM_WRITE_32_L(localbuf, xsd->nreloc); /* num of relocation entries */
-    fwrite(info->buf, 32, 1, info->f);
+    fwrite(info->buf, 40, 1, info->f);
 
     return 0;
 }
 
-static void
-xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
-		   int all_syms, /*@unused@*/ yasm_dbgfmt *df)
+static int
+xdf_objfmt_count_sym(yasm_symrec *sym, /*@null@*/ void *d)
 {
-    yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
-    xdf_objfmt_output_info info;
-    unsigned char *localbuf;
-    unsigned long symtab_count = 0;
-    unsigned long strtab_offset;
-    xdf_symtab_entry *entry;
+    /*@null@*/ xdf_objfmt_output_info *info = (xdf_objfmt_output_info *)d;
+    assert(info != NULL);
+    if (info->all_syms || yasm_symrec_get_visibility(sym) != YASM_SYM_LOCAL) {
+	/* Save index in symrec data */
+	xdf_symrec_data *sym_data = yasm_xmalloc(sizeof(xdf_symrec_data));
+	sym_data->index = info->indx;
+	yasm_symrec_add_data(sym, &xdf_symrec_data_cb, sym_data);
 
-    info.objfmt_xdf = objfmt_xdf;
-    info.f = f;
-    info.buf = yasm_xmalloc(REGULAR_OUTBUF_SIZE);
-
-    /* Allocate space for headers by seeking forward */
-    if (fseek(f, (long)(16+32*(objfmt_xdf->parse_scnum)), SEEK_SET) < 0) {
-	yasm__fatal(N_("could not seek on output file"));
-	/*@notreached@*/
-	return;
+	info->indx++;
     }
+    return 0;
+}
 
-    /* Symbol table */
-    all_syms = 1;	/* force all syms into symbol table */
-    if (all_syms) {
-	/* Need to put all local syms into XDF symbol table */
-	yasm_symtab_traverse(objfmt_xdf->symtab, objfmt_xdf,
-			     xdf_objfmt_append_local_sym);
-    }
+static int
+xdf_objfmt_output_sym(yasm_symrec *sym, /*@null@*/ void *d)
+{
+    /*@null@*/ xdf_objfmt_output_info *info = (xdf_objfmt_output_info *)d;
+    yasm_sym_vis vis = yasm_symrec_get_visibility(sym);
 
-    /* Get number of symbols */
-    if (STAILQ_EMPTY(&objfmt_xdf->xdf_symtab))
-	symtab_count = 0;
-    else {
-	/*@null@*/ /*@dependent@*/ xdf_symrec_data *sym_data_prev;
-	entry = STAILQ_LAST(&objfmt_xdf->xdf_symtab, xdf_symtab_entry, link);
-	sym_data_prev = yasm_symrec_get_data(entry->sym, &xdf_symrec_data_cb);
-	assert(sym_data_prev != NULL);
-	symtab_count = sym_data_prev->index + 1;
-    }
+    assert(info != NULL);
 
-    /* Get file offset of start of string table */
-    strtab_offset = 16+32*(objfmt_xdf->parse_scnum)+16*symtab_count;
-
-    STAILQ_FOREACH(entry, &objfmt_xdf->xdf_symtab, link) {
-	const char *name = yasm_symrec_get_name(entry->sym);
+    if (info->all_syms || vis != YASM_SYM_LOCAL) {
+	const char *name = yasm_symrec_get_name(sym);
 	const yasm_expr *equ_val;
 	const yasm_intnum *intn;
 	size_t len = strlen(name);
-	/*@dependent@*/ /*@null@*/ xdf_symrec_data *xsymd;
 	unsigned long value = 0;
 	long scnum = -3;	/* -3 = debugging symbol */
 	/*@dependent@*/ /*@null@*/ yasm_section *sect;
 	/*@dependent@*/ /*@null@*/ yasm_bytecode *precbc;
-	yasm_sym_vis vis = yasm_symrec_get_visibility(entry->sym);
 	unsigned long flags = 0;
-
-	/* Get symrec's of_data (needed for storage class) */
-	xsymd = yasm_symrec_get_data(entry->sym, &xdf_symrec_data_cb);
-	if (!xsymd)
-	    yasm_internal_error(N_("xdf: expected sym data to be present"));
+	unsigned char *localbuf;
 
 	if (vis & YASM_SYM_GLOBAL)
 	    flags = XDF_SYM_GLOBAL;
 
 	/* Look at symrec for value/scnum/etc. */
-	if (yasm_symrec_get_label(entry->sym, &precbc)) {
+	if (yasm_symrec_get_label(sym, &precbc)) {
 	    if (precbc)
 		sect = yasm_bc_get_section(precbc);
 	    else
@@ -592,7 +511,7 @@ xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
 		if (precbc)
 		    value += precbc->offset + precbc->len;
 	    }
-	} else if ((equ_val = yasm_symrec_get_equ(entry->sym))) {
+	} else if ((equ_val = yasm_symrec_get_equ(sym))) {
 	    yasm_expr *equ_val_copy = yasm_expr_copy(equ_val);
 	    intn = yasm_expr_get_intnum(&equ_val_copy,
 					yasm_common_calc_bc_dist);
@@ -613,21 +532,67 @@ xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
 	    }
 	}
 
-	localbuf = info.buf;
+	localbuf = info->buf;
 	YASM_WRITE_32_L(localbuf, scnum);	/* section number */
 	YASM_WRITE_32_L(localbuf, value);	/* value */
-	YASM_WRITE_32_L(localbuf, strtab_offset); /* string table offset */
-	strtab_offset += len+1;
+	YASM_WRITE_32_L(localbuf, info->strtab_offset);
+	info->strtab_offset += len+1;
 	YASM_WRITE_32_L(localbuf, flags);	/* flags */
-	fwrite(info.buf, 16, 1, f);
+	fwrite(info->buf, 16, 1, info->f);
+    }
+    return 0;
+}
+
+static int
+xdf_objfmt_output_str(yasm_symrec *sym, /*@null@*/ void *d)
+{
+    /*@null@*/ xdf_objfmt_output_info *info = (xdf_objfmt_output_info *)d;
+    yasm_sym_vis vis = yasm_symrec_get_visibility(sym);
+
+    assert(info != NULL);
+
+    if (info->all_syms || vis != YASM_SYM_LOCAL) {
+	const char *name = yasm_symrec_get_name(sym);
+	size_t len = strlen(name);
+	fwrite(name, len+1, 1, info->f);
+    }
+    return 0;
+}
+
+static void
+xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, int all_syms,
+		  /*@unused@*/ yasm_dbgfmt *df)
+{
+    yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
+    xdf_objfmt_output_info info;
+    unsigned char *localbuf;
+    unsigned long symtab_count = 0;
+
+    info.objfmt_xdf = objfmt_xdf;
+    info.f = f;
+    info.buf = yasm_xmalloc(REGULAR_OUTBUF_SIZE);
+
+    /* Allocate space for headers by seeking forward */
+    if (fseek(f, (long)(16+40*(objfmt_xdf->parse_scnum)), SEEK_SET) < 0) {
+	yasm__fatal(N_("could not seek on output file"));
+	/*@notreached@*/
+	return;
     }
 
-    /* String table */
-    STAILQ_FOREACH(entry, &objfmt_xdf->xdf_symtab, link) {
-	const char *name = yasm_symrec_get_name(entry->sym);
-	size_t len = strlen(name);
-	fwrite(name, len+1, 1, f);
-    }
+    /* Get number of symbols */
+    info.indx = 0;
+    info.all_syms = 1;	/* force all syms into symbol table */
+    yasm_symtab_traverse(objfmt_xdf->symtab, &info, xdf_objfmt_count_sym);
+    symtab_count = info.indx;
+
+    /* Get file offset of start of string table */
+    info.strtab_offset = 16+40*(objfmt_xdf->parse_scnum)+16*symtab_count;
+
+    /* Output symbol table */
+    yasm_symtab_traverse(objfmt_xdf->symtab, &info, xdf_objfmt_output_sym);
+
+    /* Output string table */
+    yasm_symtab_traverse(objfmt_xdf->symtab, &info, xdf_objfmt_output_str);
 
     /* Section data/relocs */
     if (yasm_object_sections_traverse(objfmt_xdf->object, &info,
@@ -646,7 +611,7 @@ xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
     YASM_WRITE_32_L(localbuf, objfmt_xdf->parse_scnum); /* number of sects */
     YASM_WRITE_32_L(localbuf, symtab_count);		/* number of symtabs */
     /* size of sect headers + symbol table + strings */
-    YASM_WRITE_32_L(localbuf, strtab_offset-16);
+    YASM_WRITE_32_L(localbuf, info.strtab_offset-16);
     fwrite(info.buf, 16, 1, f);
 
     yasm_object_sections_traverse(objfmt_xdf->object, &info,
@@ -658,19 +623,51 @@ xdf_objfmt_output(yasm_objfmt *objfmt, FILE *f, const char *obj_filename,
 static void
 xdf_objfmt_destroy(yasm_objfmt *objfmt)
 {
-    yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
-    xdf_symtab_entry *entry1, *entry2;
-
-    /* Delete local symbol table */
-    entry1 = STAILQ_FIRST(&objfmt_xdf->xdf_symtab);
-    while (entry1 != NULL) {
-	entry2 = STAILQ_NEXT(entry1, link);
-	yasm_xfree(entry1);
-	entry1 = entry2;
-    }
     yasm_xfree(objfmt);
 }
 
+static xdf_section_data *
+xdf_objfmt_init_new_section(yasm_objfmt_xdf *objfmt_xdf, yasm_section *sect,
+			    const char *sectname, unsigned long line)
+{
+    xdf_section_data *data;
+    yasm_symrec *sym;
+
+    data = yasm_xmalloc(sizeof(xdf_section_data));
+    data->scnum = objfmt_xdf->parse_scnum++;
+    data->flags = 0;
+    data->addr = NULL;
+    data->vaddr = NULL;
+    data->scnptr = 0;
+    data->size = 0;
+    data->relptr = 0;
+    data->nreloc = 0;
+    STAILQ_INIT(&data->relocs);
+    yasm_section_add_data(sect, &xdf_section_data_cb, data);
+
+    sym = yasm_symtab_define_label(objfmt_xdf->symtab, sectname,
+				   yasm_section_bcs_first(sect), 1, line);
+    data->sym = sym;
+    return data;
+}
+
+static yasm_section *
+xdf_objfmt_add_default_section(yasm_objfmt *objfmt)
+{
+    yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
+    yasm_section *retval;
+    xdf_section_data *xsd;
+    int isnew;
+
+    retval = yasm_object_get_general(objfmt_xdf->object, ".text", 0, 0, 1, 0,
+				     &isnew, 0);
+    if (isnew) {
+	xsd = xdf_objfmt_init_new_section(objfmt_xdf, retval, ".text", 0);
+	yasm_section_set_default(retval, 1);
+    }
+    return retval;
+}
+
 static /*@observer@*/ /*@null@*/ yasm_section *
 xdf_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 			    /*@unused@*/ /*@null@*/
@@ -682,11 +679,13 @@ xdf_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
     yasm_section *retval;
     int isnew;
     /*@dependent@*/ /*@null@*/ const yasm_intnum *absaddr = NULL;
-    unsigned int addralign = 0;
+    /*@dependent@*/ /*@null@*/ const yasm_intnum *vaddr = NULL;
+    unsigned long align = 0;
     unsigned long flags = 0;
     int flags_override = 0;
     char *sectname;
     int resonly = 0;
+    xdf_section_data *xsd;
 
     if (!vp || vp->param || !vp->val)
 	return NULL;
@@ -694,6 +693,12 @@ xdf_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
     sectname = vp->val;
 
     while ((vp = yasm_vps_next(vp))) {
+	if (!vp->val) {
+	    yasm__warning(YASM_WARN_GENERAL, line,
+			  N_("Unrecognized numeric qualifier"));
+	    continue;
+	}
+
 	flags_override = 1;
 	if (yasm__strcasecmp(vp->val, "use16") == 0) {
 	    flags &= ~(XDF_SECT_USE_32|XDF_SECT_USE_64);
@@ -719,30 +724,33 @@ xdf_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 			    vp->val);
 		return NULL;
 	    }
+	} else if (yasm__strcasecmp(vp->val, "virtual") == 0 && vp->param) {
+	    vaddr = yasm_expr_get_intnum(&vp->param, NULL);
+	    if (!vaddr) {
+		yasm__error(line, N_("argument to `%s' is not an integer"),
+			    vp->val);
+		return NULL;
+	    }
 	} else if (yasm__strcasecmp(vp->val, "align") == 0 && vp->param) {
-	    /*@dependent@*/ /*@null@*/ const yasm_intnum *align;
-	    unsigned long bitcnt;
+	    /*@dependent@*/ /*@null@*/ const yasm_intnum *align_expr;
 
-	    align = yasm_expr_get_intnum(&vp->param, NULL);
-	    if (!align) {
+	    align_expr = yasm_expr_get_intnum(&vp->param, NULL);
+	    if (!align_expr) {
 		yasm__error(line, N_("argument to `%s' is not a power of two"),
 			    vp->val);
 		return NULL;
 	    }
-	    addralign = yasm_intnum_get_uint(align);
+	    align = yasm_intnum_get_uint(align_expr);
 
-	    /* Check to see if alignment is a power of two.
-	     * This can be checked by seeing if only one bit is set.
-	     */
-	    BitCount(bitcnt, addralign);
-	    if (bitcnt > 1) {
+            /* Alignments must be a power of two. */
+            if ((align & (align - 1)) != 0) {
 		yasm__error(line, N_("argument to `%s' is not a power of two"),
 			    vp->val);
 		return NULL;
 	    }
 
 	    /* Check to see if alignment is supported size */
-	    if (addralign > 4096) {
+	    if (align > 4096) {
 		yasm__error(line,
 			    N_("XDF does not support alignments > 4096"));
 		return NULL;
@@ -752,33 +760,28 @@ xdf_objfmt_section_switch(yasm_objfmt *objfmt, yasm_valparamhead *valparams,
 			  N_("Unrecognized qualifier `%s'"), vp->val);
     }
 
-    retval = yasm_object_get_general(objfmt_xdf->object, sectname, 0, resonly,
-				     &isnew, line);
+    retval = yasm_object_get_general(objfmt_xdf->object, sectname, 0, align, 1,
+				     resonly, &isnew, line);
 
-    if (isnew) {
-	xdf_section_data *data;
-	yasm_symrec *sym;
+    if (isnew)
+	xsd = xdf_objfmt_init_new_section(objfmt_xdf, retval, sectname, line);
+    else
+	xsd = yasm_section_get_data(retval, &xdf_section_data_cb);
 
-	data = yasm_xmalloc(sizeof(xdf_section_data));
-	data->scnum = objfmt_xdf->parse_scnum++;
-	data->align = addralign;
-	data->flags = flags;
-	if (absaddr)
-	    data->addr = yasm_intnum_copy(absaddr);
-	else
-	    data->addr = NULL;
-	data->scnptr = 0;
-	data->size = 0;
-	data->relptr = 0;
-	data->nreloc = 0;
-	STAILQ_INIT(&data->relocs);
-	yasm_section_add_data(retval, &xdf_section_data_cb, data);
-
-	sym =
-	    yasm_symtab_define_label(objfmt_xdf->symtab, sectname,
-				     yasm_section_bcs_first(retval), 1, line);
-	xdf_objfmt_symtab_append(objfmt_xdf, sym);
-	data->sym = sym;
+    if (isnew || yasm_section_is_default(retval)) {
+	yasm_section_set_default(retval, 0);
+	xsd->flags = flags;
+	if (absaddr) {
+	    if (xsd->addr)
+		yasm_intnum_destroy(xsd->addr);
+	    xsd->addr = yasm_intnum_copy(absaddr);
+	}
+	if (vaddr) {
+	    if (xsd->vaddr)
+		yasm_intnum_destroy(xsd->vaddr);
+	    xsd->vaddr = yasm_intnum_copy(vaddr);
+	}
+	yasm_section_set_align(retval, align, line);
     } else if (flags_override)
 	yasm__warning(YASM_WARN_GENERAL, line,
 		      N_("section flags ignored on section redeclaration"));
@@ -791,6 +794,8 @@ xdf_section_data_destroy(void *data)
     xdf_section_data *xsd = (xdf_section_data *)data;
     if (xsd->addr)
 	yasm_intnum_destroy(xsd->addr);
+    if (xsd->vaddr)
+	yasm_intnum_destroy(xsd->vaddr);
     yasm_xfree(data);
 }
 
@@ -805,6 +810,8 @@ xdf_section_data_print(void *data, FILE *f, int indent_level)
     fprintf(f, "%*sflags=0x%x\n", indent_level, "", xsd->flags);
     fprintf(f, "%*saddr=", indent_level, "");
     yasm_intnum_print(xsd->addr, f);
+    fprintf(f, "%*svaddr=", indent_level, "");
+    yasm_intnum_print(xsd->vaddr, f);
     fprintf(f, "%*sscnptr=0x%lx\n", indent_level, "", xsd->scnptr);
     fprintf(f, "%*ssize=%ld\n", indent_level, "", xsd->size);
     fprintf(f, "%*srelptr=0x%lx\n", indent_level, "", xsd->relptr);
@@ -818,13 +825,9 @@ xdf_objfmt_extern_declare(yasm_objfmt *objfmt, const char *name, /*@unused@*/
 			   unsigned long line)
 {
     yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
-    yasm_symrec *sym;
 
-    sym = yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_EXTERN,
-			      line);
-    xdf_objfmt_symtab_append(objfmt_xdf, sym);
-
-    return sym;
+    return yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_EXTERN,
+			       line);
 }
 
 static yasm_symrec *
@@ -833,13 +836,9 @@ xdf_objfmt_global_declare(yasm_objfmt *objfmt, const char *name, /*@unused@*/
 			   unsigned long line)
 {
     yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
-    yasm_symrec *sym;
 
-    sym = yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_GLOBAL,
-			      line);
-    xdf_objfmt_symtab_append(objfmt_xdf, sym);
-
-    return sym;
+    return yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_GLOBAL,
+			       line);
 }
 
 static yasm_symrec *
@@ -849,14 +848,13 @@ xdf_objfmt_common_declare(yasm_objfmt *objfmt, const char *name,
 			   unsigned long line)
 {
     yasm_objfmt_xdf *objfmt_xdf = (yasm_objfmt_xdf *)objfmt;
-    yasm_symrec *sym;
 
     yasm_expr_destroy(size);
     yasm__error(line,
 	N_("XDF object format does not support common variables"));
 
-    sym = yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_COMMON, line);
-    return sym;
+    return yasm_symtab_declare(objfmt_xdf->symtab, name, YASM_SYM_COMMON,
+			       line);
 }
 
 static void
@@ -893,17 +891,16 @@ static const char *xdf_objfmt_dbgfmt_keywords[] = {
 
 /* Define objfmt structure -- see objfmt.h for details */
 yasm_objfmt_module yasm_xdf_LTX_objfmt = {
-    YASM_OBJFMT_VERSION,
     "Extended Dynamic Object",
     "xdf",
     "xdf",
-    ".text",
     32,
     xdf_objfmt_dbgfmt_keywords,
     "null",
     xdf_objfmt_create,
     xdf_objfmt_output,
     xdf_objfmt_destroy,
+    xdf_objfmt_add_default_section,
     xdf_objfmt_section_switch,
     xdf_objfmt_extern_declare,
     xdf_objfmt_global_declare,
