@@ -25,7 +25,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: nasm-parser.c 1152 2004-10-02 06:18:30Z peter $");
+/*@unused@*/ RCSID("$Id: nasm-parser.c 1591 2006-07-13 03:25:38Z peter $");
 
 #define YASM_LIB_INTERNAL
 #include <libyasm.h>
@@ -35,8 +35,9 @@
 
 static void
 nasm_parser_do_parse(yasm_object *object, yasm_preproc *pp, yasm_arch *a,
-		     yasm_objfmt *of, FILE *f, const char *in_filename,
-		     int save_input, yasm_section *def_sect)
+		     yasm_objfmt *of, yasm_dbgfmt *df, FILE *f,
+		     const char *in_filename, int save_input,
+		     yasm_section *def_sect)
 {
     yasm_parser_nasm parser_nasm;
 
@@ -52,6 +53,7 @@ nasm_parser_do_parse(yasm_object *object, yasm_preproc *pp, yasm_arch *a,
     parser_nasm.preproc = pp;
     parser_nasm.arch = a;
     parser_nasm.objfmt = of;
+    parser_nasm.dbgfmt = df;
 
     parser_nasm.cur_section = def_sect;
     parser_nasm.prev_bc = yasm_section_bcs_first(def_sect);
@@ -95,7 +97,6 @@ static const char *nasm_parser_preproc_keywords[] = {
 
 /* Define parser structure -- see parser.h for details */
 yasm_parser_module yasm_nasm_LTX_parser = {
-    YASM_PARSER_VERSION,
     "NASM-compatible parser",
     "nasm",
     nasm_parser_preproc_keywords,
