@@ -1,4 +1,4 @@
-/* $Id: nasm-parser.h 1152 2004-10-02 06:18:30Z peter $
+/* $Id: nasm-parser.h 1591 2006-07-13 03:25:38Z peter $
  * NASM-compatible parser header file
  *
  *  Copyright (C) 2002  Peter Johnson
@@ -29,7 +29,7 @@
 
 #include "nasm-bison.h"
 
-#define YYCTYPE		char
+#define YYCTYPE		unsigned char
 typedef struct Scanner {
     YYCTYPE		*bot, *tok, *ptr, *cur, *pos, *lim, *top, *eof;
     unsigned int	tchar, tline, cline;
@@ -51,6 +51,7 @@ typedef struct yasm_parser_nasm {
     /*@dependent@*/ yasm_preproc *preproc;
     /*@dependent@*/ yasm_arch *arch;
     /*@dependent@*/ yasm_objfmt *objfmt;
+    /*@dependent@*/ yasm_dbgfmt *dbgfmt;
 
     /*@dependent@*/ yasm_linemap *linemap;
     /*@dependent@*/ yasm_symtab *symtab;
@@ -68,7 +69,8 @@ typedef struct yasm_parser_nasm {
 	DIRECTIVE,
 	DIRECTIVE2,
 	LINECHG,
-	LINECHG2
+	LINECHG2,
+	INSTRUCTION
     } state;
 } yasm_parser_nasm;
 
