@@ -1,4 +1,4 @@
-/* $Id: util.h 1148 2004-09-25 06:38:57Z peter $
+/* $Id: util.h 1591 2006-07-13 03:25:38Z peter $
  * YASM utility functions.
  *
  * Includes standard headers and defines prototypes for replacement functions
@@ -102,6 +102,9 @@
 #elif HAVE_STRICMP
 # define yasm__strcasecmp(x, y)		stricmp(x, y)
 # define yasm__strncasecmp(x, y, n)	strnicmp(x, y, n)
+#elif HAVE__STRICMP
+# define yasm__strcasecmp(x, y)		_stricmp(x, y)
+# define yasm__strncasecmp(x, y, n)	_strnicmp(x, y, n)
 #elif HAVE_STRCMPI
 # define yasm__strcasecmp(x, y)		strcmpi(x, y)
 # define yasm__strncasecmp(x, y, n)	strncmpi(x, y, n)
